"""
This file contains helper functions to save a linear programming relaxation
problem to file.

@authors: Erica Han, Elie Wolfe
"""

import mosek
from copy import deepcopy


def write_to_lp(args: dict, filename: str) -> None:
    """Export the problem to a file in .lp format. Specification can be found
    at https://docs.mosek.com/latest/pythonapi/lp-format.html.

    Parameters
    ----------
    args : dict
        The arguments of the problem to write to the file.
    filename : str
        The file to write to.
    """
    processed_args = process_arguments(args)
    objective = processed_args["objective"]
    known_vars = processed_args["known_vars"]
    internal_equalities = processed_args["equalities"]
    inequalities = processed_args["inequalities"]
    lower_bounds = processed_args["lower_bounds"]
    upper_bounds = processed_args["upper_bounds"]
    variables = processed_args["variables"]


    # Write problem to file
    f = open(filename, "w")
    f.write(f"\\ File: {filename}\n")
    f.write("maximize\n")
    f.write(f"obj: {format_constraint_lp(known_vars, objective, 'obj')}\n")
    f.write("subject to\n")

    i = 0
    for eq in internal_equalities:
        cons = format_constraint_lp(known_vars, eq, 'eq')
        if cons != "":
            f.write(f"c{i}: {cons}\n")
            i += 1
    for ineq in inequalities:
        cons = format_constraint_lp(known_vars, ineq, 'ineq')
        if cons != "":
            f.write(f"c{i}: {cons}\n")
            i += 1

    f.write("bounds\n")
    for x in lower_bounds:
        f.write(f"{lower_bounds[x]} <= {x}")
    for x in upper_bounds:
        f.write(f"{x} <= {upper_bounds[x]}")
    f.write("end")
    f.close()


def write_to_mps(args: dict, filename: str) -> None:
    """Export the problem to a file in .mps format. Specification can be found
    at https://docs.mosek.com/latest/pythonapi/mps-format.html.

    Parameters
    ----------
    args : dict
        The arguments of the problem to write to the file.
    filename : str
        The file to write to.
    """
    processed_args = process_arguments(args)
    objective = processed_args["objective"]
    known_vars = processed_args["known_vars"]
    internal_equalities = processed_args["equalities"]
    inequalities = processed_args["inequalities"]
    lower_bounds = processed_args["lower_bounds"]
    upper_bounds = processed_args["upper_bounds"]
    variables = processed_args["variables"]

    # Exclude constraints in which variables are all known values
    internal_equalities = [eq for eq in internal_equalities
                           if set(eq).difference(known_vars)]
    inequalities = [ineq for ineq in inequalities
                    if set(ineq).difference(known_vars)]
    constraints = internal_equalities + inequalities
    nof_equalities = len(internal_equalities)
    nof_inequalities = len(inequalities)

    f = open(filename, "w")
    f.write(f"* File: {filename}\n")
    f.write(f"NAME          {filename.split('.')[0]}\n")
    f.write(f"OBJSENSE\n    MAX\n")

    # Specify rows of the problem
    f.write(f"ROWS\n")
    f.write(f" N  obj\n")
    for i in range(nof_equalities):
        f.write(f" E  c{i}\n")
    for i in range(nof_inequalities):
        f.write(f" G  c{nof_equalities + i}\n")

    # Specify columns of the problem
    f.write(f"COLUMNS\n")
    for x in variables:
        if x in set(objective).difference(known_vars):
            arr = [x, 'obj', objective[x]]
            f.write("    {:10}{:10}{:<10.4f}\n".format(*arr))
        for i, eq in enumerate(internal_equalities):
            if x in set(eq).difference(known_vars):
                arr = [x, f"c{i}", eq[x]]
                f.write("    {:10}{:10}{:<10.4f}\n".format(*arr))
        for i, ineq in enumerate(inequalities):
            if x in set(ineq).difference(known_vars):
                arr = [x, f"c{nof_equalities + i}",
                       ineq[x]]
                f.write("    {:10}{:10}{:<10.4f}\n".format(*arr))

    f.write(f"RHS\n")
    # Specify objective offset
    rhs = 0
    for x in set(objective).intersection(known_vars):
        rhs -= objective[x] * known_vars[x]
    if rhs != 0:
        arr = ['rhs', 'obj', rhs]
        f.write("    {:10}{:10}{:<10.4f}\n".format(*arr))

    # Specify non-zero RHS values of the constraints
    for i, c in enumerate(constraints):
        rhs = 0
        for x in set(c).intersection(known_vars):
            rhs -= c[x] * known_vars[x]
        if rhs != 0:
            arr = ['rhs', f"c{i}", rhs]
            f.write("    {:10}{:10}{:<10.4f}\n".format(*arr))

    f.write(f"RANGES\n")

    # Specify variable bound values
    f.write(f"BOUNDS\n")
    for x in lower_bounds:
        arr = ['bound', x, lower_bounds[x]]
        f.write(" LO {:10}{:10}{:<10.4f}\n".format(*arr))
    for x in upper_bounds:
        arr = ['bound', x, upper_bounds[x]]
        f.write(" UP {:10}{:10}{:<10.4f}\n".format(*arr))
    f.write(f"ENDATA")
    f.close()


def lp_to_mps(args: dict, filename: str) -> None:
    """Export the problem to a file in .mps format by writing it as an LP file
    then converting it to MPS through Mosek.

    Parameters
    ----------
    args : dict
        The arguments of the problem to write to the file.
    filename : str
        The file to write to.
    """
    lp_filename = f"{filename.split('.')[0]}.lp"
    write_to_lp(args, lp_filename)
    with mosek.Task() as t:
        t.readdata(f"{lp_filename}")
        t.writedata(filename)


def process_arguments(args: dict) -> dict:
    """Helper function that processes arguments and variables in preparation
    for writing.

    Parameters
    ----------
    args : dict
        The arguments to be processed.

    Returns
    -------
    dict
        Processed objective, known_vars, internal_equalities,
        inequalities, lower_bounds, upper_bounds, variables
    """
    old_variables = args["variables"]
    old_equalities = deepcopy(args["equalities"])
    old_semiknown = args["semiknown_vars"]
    for x, (c, x2) in old_semiknown.items():
        old_equalities.append({x: 1, x2: -c})
    old_inequalities = args["inequalities"]
    old_known_vars = args["known_vars"]
    old_objective = args.get("objective", {})
    old_lower_bounds = args.get("lower_bounds", {})
    old_upper_bounds = args.get("upper_bounds", {})
    var_idx =  {x: f"x{i}" for i, x in enumerate(old_variables)}
    new_variables = list(var_idx.values())
    new_equalities = [{var_idx[x]: c for x, c in eq.items()} for eq in  old_equalities]
    new_inequalities = [{var_idx[x]: c for x, c in ineq.items()} for ineq in old_inequalities]
    new_known_vars = {var_idx[x]: c for x, c in old_known_vars.items()}
    new_lower_bounds = {var_idx[x]: c for x, c in old_lower_bounds.items()}
    new_upper_bounds = {var_idx[x]: c for x, c in old_upper_bounds.items()}
    new_objective = {var_idx[x]: c for x, c in old_objective.items()}

    return {
        "objective": new_objective,
        "variables": new_variables,
        "equalities": new_equalities,
        "inequalities": new_inequalities,
        "known_vars": new_known_vars,
        "lower_bounds": new_lower_bounds,
        "upper_bounds": new_upper_bounds}


def format_constraint_lp(known_vars: dict,
                         constraint: dict,
                         cons_type: str) -> str:
    """Helper function to format constraints (and the objective) for the LP
    file format.

    Parameters
    ----------
    known_vars : dict
        Known values of variables.
    constraint : dict
        The constraint to be formatted.
    cons_type : str
        Whether the constraint is an objective ('obj'), equality ('eq'), or
        inequality ('ineq).

    Returns
    -------
    str
        The constraint as a human-readable string
    """
    # Process known values
    c = 0
    for x in set(constraint).intersection(known_vars):
        c += constraint[x] * known_vars[x]

    # Process unknown values
    cons = ""
    for x in set(constraint).difference(known_vars):
        cons += f"{constraint[x]} {x} + "
    if cons == "":
        return cons

    if cons_type == 'obj':
        cons += str(c)
    elif cons_type == 'eq':
        cons = cons[:-3] + ' = ' + str(-c)
    elif cons_type == 'ineq':
        cons = cons[:-3] + ' >= ' + str(-c)
    return cons



import unittest
import warnings
import mosek
import os
import numpy as np

from inflation import InflationProblem, InflationLP, InflationSDP
from inflation.lp.writer_utils import write_to_lp, write_to_mps
from inflation.sdp.writer_utils import write_to_csv, write_to_mat, write_to_sdpa
from scipy.io import loadmat


class TestLPWriters(unittest.TestCase):
    p = np.zeros((2, 2, 2, 1))
    p[0, 0, 0, 0] = 0.3
    p[1, 0, 0, 0] = 0.7
    p[0, 0, 1, 0] = 0.7
    p[1, 1, 1, 0] = 0.3
    p = 0.9 * p + 0.1 * (1 / 4)

    instrumental = InflationProblem({"U_AB": ["A", "B"],
                                     "A": ["B"]},
                                    outcomes_per_party=(2, 2),
                                    settings_per_party=(2, 1),
                                    order=("A", "B"),
                                    classical_sources=["U_AB"])
    instrumental_infLP = InflationLP(instrumental,
                                     verbose=False,
                                     include_all_outcomes=True)
    instrumental_infLP.set_distribution(p)
    instrumental_infLP.set_objective(objective={'P[B=1|do(A=0)]': 1},
                                     direction='max')
    instrumental_infLP.solve()
    args = instrumental_infLP._prepare_solver_arguments(separate_bounds=True)
    primal_value = instrumental_infLP.objective_value

    def test_write_to_lp(self):
        self.ext = 'lp'
        write_to_lp(self.args, "inst.lp")
        with mosek.Task() as task:
            task.readdata("inst.lp")
            task.optimize()
            obj = task.getprimalobj(mosek.soltype.bas)
        self.assertAlmostEqual(self.primal_value, obj,
                               msg="The expected value and value when reading "
                                   "from LP are not equal.")

    def test_write_to_mps(self):
        self.ext = 'mps'
        write_to_mps(self.args, "inst.mps")
        with mosek.Task() as task:
            task.readdata("inst.mps")
            task.optimize()
            obj = task.getprimalobj(mosek.soltype.bas)
        self.assertAlmostEqual(self.primal_value, obj,
                               msg="The expected value and value when reading "
                                   "from MPS are not equal.")

    def tearDown(self):
        os.remove(f"inst.{self.ext}")

class TestSDPWriters(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        warnings.simplefilter("ignore", category=UserWarning)

    # Create an example problem
    scenario = InflationProblem({"Lambda": ["A"]},
                                outcomes_per_party=[3],
                                settings_per_party=[2],
                                inflation_level_per_source=[2])
    sdp = InflationSDP(scenario)
    sdp.generate_relaxation("npa3")
    
    c1 = 0.1
    c2 = -0.2
    v  = 0.5
    ub = 0.4
    lb = 0.3
    sdp.set_objective({"P[A_0=0]": c1, "P[A_1=1]": c2}, "max")
    sdp.set_bounds({"P[A_0=1]": ub})
    sdp.set_bounds({"P[A_0=1]": 0.3}, "lo")
    sdp.set_values({"P[A_1=0]": v}, use_lpi_constraints=True)
    sdp.set_extra_equalities([{"P[A_0=0]": c2, "P[A_0=1]": -c1}])
    sdp.set_extra_inequalities([{"P[A_0=0]": c1, "P[A_0=1]": -c2}])

    def test_write_to_csv(self):
        self.ext = 'csv'
        # Write the problem to a file
        write_to_csv(self.sdp, 'inst.csv')

        # Read the contents of the file
        with open('inst.csv', 'r') as file:
            contents = file.read()

        # Assert that the file contains the objective function
        self.assertIn(f"Objective: {self.c1}*P[A_0=0]{self.c2}*P[A_1=1]",
                      contents,
                      "The objective function is not exported.")

        # Assert that the file contains the variable constraints
        self.assertEqual(float(contents.split("\n")[1].split(",")[3]), self.v,
                        "The variable constraints are not implemented/correct.")
        
        # Assert that the file contains LPI constraints
        self.assertEqual(contents.split("\n")[1].split(",")[13],
                         f"{self.v}*P[A_0=0]",
                         "The LPI constraints are not implemented/correct.")

        # Checks on bounds
        bounds = contents.split("\n")[140:308]
        lbs = list(float(b.split(",")[1]) for b in bounds)
        ubs = [b.split(",")[2] for b in bounds]
        
        # Assert that the file contains the upper bounds
        self.assertIn(str(self.ub), ubs,
                      "The upper bounds are not implemented/correct.")

        # Assert that the file contains the lower bounds
        self.assertTrue(sum(lbs) == self.lb,
                        "The lower bounds are not implemented/correct.")

        # Assert that the file contains the moment equalities
        self.assertIn(f"{self.c2}*P[A_0=0]-{self.c1}*P[A_0=1]", contents,
                      "The moment equalities are not implemented.")
        
        # Assert that the file contains the moment inequalities
        self.assertIn(f"{self.c1}*P[A_0=0]+{abs(self.c2)}*P[A_0=1]", contents,
                      "The moment inequalities are not implemented.")
        
        # ineq = moment_inequalities[0][0][0][0]
        # self.assertTrue(np.array_equal(ineq[0], [[3, 4]])
        #                 and np.array_equal(ineq[1], [[self.c1, -self.c2]]),
        #                 "The moment inequalities are not correct.")
        
    def test_write_to_mat(self):
        self.ext = 'mat'
        # Write the problem to a file
        write_to_mat(self.sdp, 'inst.mat')

        # Read the contents of the file
        contents = loadmat('inst.mat')

        moments_idx2name    = contents['moments_idx2name']
        momentmatrix        = contents['momentmatrix']
        known_moments       = contents['known_moments']
        semiknown_moments   = contents['semiknown_moments']
        objective           = contents['objective']
        moment_lowerbounds  = contents['moment_lowerbounds']
        moment_upperbounds  = contents['moment_upperbounds']
        moment_equalities   = contents['moment_equalities']
        moment_inequalities = contents['moment_inequalities']

        # Assert that the file contains the expected number of variables
        nvars = (len(moments_idx2name) - len(known_moments)
                 - len(semiknown_moments))
        self.assertEqual(nvars, 661,
                         "The number of variables is not correct.")

        # Assert that the file contains the objective function
        self.assertTrue(np.array_equal(objective, [[3, self.c1], [6, self.c2]]),
                        "The objective function is not correct.")

        # Assert that the file contains the variable constraints
        self.assertTrue(np.array_equal(known_moments, [[1, 0],
                                                       [2, 1],
                                                       [5, self.v],
                                                       [18, self.v**2]]),
                        "The variable constraints are not implemented/correct.")
        
        # Assert that the file contains LPI constraints
        self.assertTrue(all(semiknown_moments[:,1] == self.v),
                        "The LPI constraints are not implemented/correct.")

        # Assert that the file contains the upper bounds
        self.assertTrue(np.array_equal(moment_upperbounds, [[4, self.ub]]),
                        "The upper bounds are not implemented/correct.")

        # Assert that the file contains the lower bounds
        self.assertTrue(len(moment_lowerbounds) == 168
                        and np.array_equal(moment_lowerbounds[1], [4, self.lb])
                        and sum(moment_lowerbounds[:, 1]) == self.lb,
                        "The lower bounds are not implemented/correct.")

        # Assert that the file contains the moment equalities
        self.assertEqual(len(moment_equalities), 1,
                         "The moment equalities are not implemented.")
        
        eq = moment_equalities[0][0][0][0]
        self.assertTrue(np.array_equal(eq[0], [[3, 4]])
                        and np.array_equal(eq[1], [[self.c2, -self.c1]]),
                        "The moment equalities are not correct.")

        # Assert that the file contains the moment inequalities
        self.assertEqual(len(moment_inequalities), 1,
                         "The moment inequalities are not implemented.")
        
        ineq = moment_inequalities[0][0][0][0]
        self.assertTrue(np.array_equal(ineq[0], [[3, 4]])
                        and np.array_equal(ineq[1], [[self.c1, -self.c2]]),
                        "The moment inequalities are not correct.")
        
    def test_write_to_sdpa(self):
        self.ext = 'dat-s'
        # Write the problem to a file
        write_to_sdpa(self.sdp, 'inst.dat-s')

        # Read the contents of the file
        with open('inst.dat-s', "r") as file:
            contents = file.read()

        # Assert that the file is not empty
        self.assertNotEqual(contents, "")

        # Assert that the file contains the expected number of variables and blocks
        vars, blocks, struct = contents.split("\n")[1:4]
        self.assertTrue(vars == "661 = number of vars",
                        "The number of variables is not correct.")
        self.assertTrue(blocks == "5 = number of blocks",
                        "The number of blocks is not correct.")
        self.assertTrue(struct == "(137,-1,-141,-2,-1) = BlockStructure",
                        "The block structure is not correct.")

        # Assert that the file contains the objective function
        objective = contents.split("\n")[4]
        self.assertTrue(objective[0] == '{' and objective[-1] == '}',
                        "The objective function is not created.")
        self.assertTrue(float(objective.split(",")[0][1:]) == self.c1
                        and float(objective.split(",")[2]) == self.c2,
                        "The objective function is not correct.")

        # Assert that the file contains the variable constraints
        self.assertIn(f"0\t1\t1\t4\t-{self.v}", contents,
                      "The variable constraints are not implemented/correct.")
        
        # Assert that the file contains LPI constraints
        self.assertIn(f"1\t1\t1\t14\t{self.v}", contents,
                      "The LPI constraints are not implemented/correct.")

        # Assert that the file contains the upper bounds
        self.assertTrue(f"0\t2\t1\t1\t-{self.ub}" in contents
                        and f"2\t2\t1\t1\t-1.0" in contents,
                        "The upper bounds are not implemented/correct.")

        # Assert that the file contains the lower bounds
        self.assertTrue(f"0\t3\t2\t2\t{self.lb}" in contents
                        and f"2\t3\t2\t2\t1.0" in contents,
                        "The lower bounds are not implemented/correct.")

        # Assert that the file contains the moment equalities
        self.assertIn(f"1\t4\t1\t1\t{self.c2}\n"
                      + f"1\t4\t2\t2\t{-self.c2}\n"
                      + f"2\t4\t1\t1\t{-self.c1}\n"
                      + f"2\t4\t2\t2\t{self.c1}", contents,
                      "The moment equalities are not implemented/correct.")

        # Assert that the file contains the moment inequalities
        self.assertIn(f"1\t5\t1\t1\t{self.c1}\n2\t5\t1\t1\t{-self.c2}\n",
                      contents,
                      "The moment inequalities are not implemented/correct.")

    def tearDown(self):
        os.remove(f"inst.{self.ext}")



import unittest
import numpy as np
import warnings
from scipy.sparse import lil_matrix, coo_matrix, vstack
from copy import deepcopy

from inflation.sdp.sdp_utils import solveSDP_MosekFUSION
from inflation.lp.lp_utils import solveLP_sparse, to_sparse, convert_dicts, \
    solveLP

simple_lp = {
    "objective": {'x': 1, 'y': 1, 'z': 1, 'w': -2},  # x + y + z - 2w
    "known_vars": {'1': 1},  # Define the variable that is the identity
    "inequalities": [{'x': -1, '1': 2},  # 2 - x >= 0
                     {'y': -1, '1': 5},  # 5 - y >= 0
                     {'z': -1, '1': 1 / 2},  # 1/2 - z >= 0
                     {'w': 1, '1': 1}],  # w >= -1
    "equalities": [{'x': 1 / 2, 'y': 2, '1': -3}]  # x/2 + 2y - 3 = 0
}
var = ['1', 'w', 'x', 'y', 'z']
simple_lp_mat = convert_dicts(**simple_lp, variables=var)


class TestSDP(unittest.TestCase):
    G = np.array([[1,  2,  3,  4,  5],
                  [2,  1,  6,  7,  8],
                  [3,  6,  1,  9, 10],
                  [4,  7,  9,  1, 11],
                  [5,  8, 10, 11,  1]])
    mask_matrices = {}
    for i in np.unique(G):
        mask_matrices.update({str(i): lil_matrix(G == i)})
    simple_sdp = {"mask_matrices": mask_matrices,
                  "objective":  {'7': 1, '8': 1, '9': 1, '10': -1},
                  "known_vars": {'1': 1}
                  }

    def test_LP_with_SDP(self):
        primal_sol   = solveSDP_MosekFUSION(**simple_lp, solve_dual=False)
        dual_sol     = solveSDP_MosekFUSION(**simple_lp, solve_dual=True)
        value_primal = primal_sol["primal_value"]
        value_dual   = dual_sol["primal_value"]
        self.assertTrue(np.isclose(value_primal, value_dual),
                        "The dual and primal solutions in LP are not equal.")
        self.assertTrue(np.isclose(value_dual, 2 + 1 + 1/2 + 2),
                        "The solution to a simple LP is not correct.")

    def test_SDP(self):
        # Maximization of CHSH on NPA level 1+AB
        primal_sol   = solveSDP_MosekFUSION(**self.simple_sdp,
                                            solve_dual=False)
        dual_sol     = solveSDP_MosekFUSION(**self.simple_sdp, solve_dual=True)
        value_primal = primal_sol["primal_value"]
        value_dual   = dual_sol["primal_value"]
        self.assertTrue(np.isclose(value_primal, value_dual, atol=1e-5),
                        "The dual and primal solutions in SDP are not equal.")
        self.assertTrue(np.isclose(value_primal, 2*np.sqrt(2)),
                        "The solution to a simple SDP is not correct")

    def test_SDP_equalities(self):
        # Maximize CHSH while the two-body correlators satisfy a local model
        import sympy as sp
        q = np.zeros((4, 4), dtype=object)
        for i, j in np.ndindex(4, 4):
            q[i, j] = sp.Symbol(f'q{i}{j}')

        A0B0 = (q[0, 0] - q[0, 1] + q[0, 2] - q[0, 3] - q[1, 0] + q[1, 1] -
                q[1, 2] + q[1, 3] + q[2, 0] - q[2, 1] + q[2, 2] - q[2, 3] -
                q[3, 0] + q[3, 1] - q[3, 2] + q[3, 3])
        A0B1 = (q[0, 0] + q[0, 1] - q[0, 2] - q[0, 3] - q[1, 0] - q[1, 1] +
                q[1, 2] + q[1, 3] + q[2, 0] + q[2, 1] - q[2, 2] - q[2, 3] -
                q[3, 0] - q[3, 1] + q[3, 2] + q[3, 3])
        A1B0 = (q[0, 0] - q[0, 1] + q[0, 2] - q[0, 3] + q[1, 0] - q[1, 1] +
                q[1, 2] - q[1, 3] - q[2, 0] + q[2, 1] - q[2, 2] + q[2, 3] -
                q[3, 0] + q[3, 1] - q[3, 2] + q[3, 3])
        A1B1 = (q[0, 0] + q[0, 1] - q[0, 2] - q[0, 3] + q[1, 0] + q[1, 1] -
                q[1, 2] - q[1, 3] - q[2, 0] - q[2, 1] + q[2, 2] + q[2, 3] -
                q[3, 0] - q[3, 1] + q[3, 2] + q[3, 3])
        problem = {**self.simple_sdp,
                   **{"inequalities": [*[{q[idx]: 1}    # Positivity
                                       for idx in np.ndindex(4, 4)]],
                      "equalities": [
                        # Normalization
                        {**{q[idx]: 1 for idx in np.ndindex(4, 4)}, '1': -1},
                        # LHV
                        {**A0B0.as_coefficients_dict(), '7': -1},
                        {**A0B1.as_coefficients_dict(), '8': -1},
                        {**A1B0.as_coefficients_dict(), '9': -1},
                        {**A1B1.as_coefficients_dict(), '10': -1}
                                        ]
                      }
                   }
        primal_sol   = solveSDP_MosekFUSION(**problem, solve_dual=False)
        dual_sol     = solveSDP_MosekFUSION(**problem, solve_dual=True)
        value_primal = primal_sol["primal_value"]
        value_dual   = dual_sol["primal_value"]
        self.assertTrue(np.isclose(value_primal, value_dual),
                        "The dual and primal solutions in SDPs with " +
                        "additional variables are not equal.")
        self.assertTrue(np.isclose(value_primal, 2),
                        "Max CHSH over local strategies is not 2.")
        # Check that some of the constraints are satisfied
        vals = dual_sol['x']
        for i in range(4):
            for j in range(4):
                self.assertTrue(vals[q[i, j]] >= -1e-9,
                                f"q[{i}, {j}] is negative.")
        self.assertTrue(np.isclose(1,
                                   sum([vals[q[idx]]
                                        for idx in np.ndindex(4, 4)])),
                        "The local model is not normalized.")

    def test_SDP_inequalities(self):
        # Maximize CHSH on NPA level 1+AB subject to CHSH <= bound
        bound = 2.23
        problem = {**self.simple_sdp,
                   **{"inequalities":
                      [{'1': bound, '7': -1, '8': -1, '9': -1, '10': 1}]}
                   }
        primal_sol   = solveSDP_MosekFUSION(**problem, solve_dual=False)
        dual_sol     = solveSDP_MosekFUSION(**problem, solve_dual=True)
        value_primal = primal_sol["primal_value"]
        value_dual   = dual_sol["primal_value"]
        self.assertTrue(np.isclose(value_primal, value_dual),
                        "The dual and primal solutions in bounded SDPs are " +
                        "not equal.")
        self.assertTrue(np.isclose(value_primal, bound),
                        f"Max CHSH with CHSH <= {bound} is not {bound}.")


class TestLP(unittest.TestCase):
    lp_expected_sol_free_bounds = {
        "primal_value": 5.5,
        "dual_certificate": {'1': 5.5, 'w': 2, 'x': -1, 'y': -1, 'z': -1},
        "x": {'x': 2, 'y': 1, 'z': 1 / 2, 'w': -1, '1': 1}
    }
    lp_expected_sol_non_neg_bounds = {
        "primal_value": 3.5,
        "dual_certificate": {'1': 3.5, 'w': 2, 'x': -1, 'y': -1, 'z': -1},
        "x": {'x': 2.0, 'y': 1.0, 'w': 0.0, 'z': 0.5, '1': 1}
    }

    def test_LP_free_bounds(self):
        expected_sol = self.lp_expected_sol_free_bounds
        actual_sols = self.setup_LP_test_case({"default_non_negative": False})
        self.check_solution(expected_sol, actual_sols)

    def test_LP_non_negative_bounds(self):
        expected_sol = self.lp_expected_sol_non_neg_bounds
        actual_sols = self.setup_LP_test_case({"default_non_negative": True})
        self.check_solution(expected_sol, actual_sols)

    def test_LP_lower_bounds_of_zero(self):
        expected_sol = self.lp_expected_sol_non_neg_bounds
        actual_sols = self.setup_LP_test_case({
            "lower_bounds": {'x': 0, 'y': 0, 'z': 0, 'w': 0},
            "default_non_negative": False})
        self.check_solution(expected_sol, actual_sols)

    def test_LP_lower_bounds_of_zero_and_non_negative(self):
        expected_sol = self.lp_expected_sol_non_neg_bounds
        actual_sols = self.setup_LP_test_case({
            "lower_bounds": {'x': 0, 'y': 0, 'z': 0, 'w': 0},
            "default_non_negative": True})
        self.check_solution(expected_sol, actual_sols)

    def test_LP_negative_lower_bounds(self):
        expected_sol = {
            "primal_value": 0.0,
            "dual_certificate": {'1': 8.0, 'w': 2, 'x': -1, 'y': -1, 'z': -1},
            "x": {'x': -2.0, 'y': 2.0, 'z': -2.0, 'w': -1.0, '1': 1}}
        actual_sols = self.setup_LP_test_case({
            "lower_bounds": {'x': -2, 'y': 2},
            "upper_bounds": {'z': -2, 'w': 2},
            "default_non_negative": False})
        self.check_solution(expected_sol, actual_sols)

    def test_LP_negative_lower_bounds_and_non_negative(self):
        expected_sol = self.lp_expected_sol_non_neg_bounds
        actual_sols = self.setup_LP_test_case({
            "lower_bounds": {'x': -1, 'y': 1},
            "upper_bounds": {'w': 1},
            "default_non_negative": True})
        self.check_solution(expected_sol, actual_sols)

    @staticmethod
    def setup_LP_test_case(args):
        """Given problem arguments, set up dictionary of solutions from the
        solver."""
        args.update(convert_dicts(**args, variables=var))
        mat_primal_sol = solveLP_sparse(**simple_lp_mat, **args,
                                        variables=var, solve_dual=False)
        mat_dual_sol = solveLP_sparse(**simple_lp_mat, **args,
                                      variables=var, solve_dual=True)
        actual_sols = {
            "sparse, primal": mat_primal_sol,
            "sparse, dual": mat_dual_sol
        }
        return actual_sols

    def check_solution(self, exp_sol, act_sols):
        """Asserts that primal values, certificates, and x values are all
        correct."""
        primal_values = {prob: sol["primal_value"]
                         for prob, sol in act_sols.items()}
        for prob, value in primal_values.items():
            with self.subTest(msg="Testing primal values", i=prob):
                self.assertEqual(exp_sol["primal_value"], value,
                                 f"The objective value is incorrect.")
        certificates = {prob: sol["dual_certificate"]
                        for prob, sol in act_sols.items()}
        for prob, cert in certificates.items():
            with self.subTest(msg="Testing certificates", i=prob):
                self.assertEqual(exp_sol["dual_certificate"], cert,
                                 f"The dual certificate is incorrect.")
        x_values = {prob: sol["x"] for prob, sol in act_sols.items()}
        for prob, x in x_values.items():
            with self.subTest(msg="Testing x values", i=prob):
                self.assertEqual(exp_sol["x"], x,
                                 f"The solution values are incorrect.")


class TestSolverProcesses(unittest.TestCase):
    def test_semiknown_constraints(self):
        """Check that semiknown_moments are correctly processed."""
        problem = {
            "objective":  {'x': -1, 'y': -2, 'z': -1},
            "known_vars": {'1': 1},
            "inequalities":  [{'x': 1, 'y': 1, 'z': 1, '1': -26},
                              {'x': 1, '1': -3},
                              {'y': 1, '1': -4},
                              {'z': 1, '1': -1}],
            "equalities": [{'x': -5, 'y': 1, 'z': -2, '1': -7}]
        }
        p = solveSDP_MosekFUSION(**problem,
                                 semiknown_vars={},
                                 solve_dual=False,
                                 process_constraints=False)
        p_lpi = solveSDP_MosekFUSION(**problem,
                                     semiknown_vars={'z': (0.5, 'x')},
                                     solve_dual=False,
                                     process_constraints=False)
        p_lpi_process = solveSDP_MosekFUSION(**problem,
                                             semiknown_vars={'z': (0.5, 'x')},
                                             solve_dual=False,
                                             process_constraints=True)
        d = solveSDP_MosekFUSION(**problem,
                                 semiknown_vars={},
                                 solve_dual=True,
                                 process_constraints=False)
        d_lpi = solveSDP_MosekFUSION(**problem,
                                     semiknown_vars={'z': (0.5, 'x')},
                                     solve_dual=True,
                                     process_constraints=False)
        d_lpi_process = solveSDP_MosekFUSION(**problem,
                                             semiknown_vars={'z': (0.5, 'x')},
                                             solve_dual=True,
                                             process_constraints=True)
        p_lp = solveLP(**problem, 
                       semiknown_vars={},
                       solve_dual=False)
        p_lpi_lp = solveLP(**problem,
                           semiknown_vars={'z': (0.5, 'x')}, 
                           solve_dual=False)
        d_lp = solveLP(**problem,
                       semiknown_vars={},
                       solve_dual=True)
        d_lpi_lp = solveLP(**problem,
                           semiknown_vars={'z': (0.5, 'x')},
                           solve_dual=True)

        truth_obj, truth_obj_lpi = -52, -109/2
        truth_x =     {'x': 3, 'y': 24, 'z': 1, '1': 1}
        truth_x_lpi = {'x': 3, 'y': 25, 'z': 3/2, '1': 1}

        msg = "The dual and primal solutions are not equal when " + \
              "processing semiknown constraints."

        check = lambda x: np.isclose(x, truth_obj)
        self.assertTrue(all(map(check, [p['primal_value'],
                                        d['primal_value'],
                                        p_lp['primal_value'],
                                        d_lp['primal_value'],
                                        p['dual_value'],
                                        d['dual_value'],
                                        p_lp['dual_value'],
                                        d_lp['dual_value'],
                                        ])), msg)

        check = lambda x: np.isclose(x, truth_obj_lpi)
        self.assertTrue(all(map(check, [p_lpi['primal_value'],
                                        d_lpi['primal_value'],
                                        p_lpi_process['primal_value'],
                                        d_lpi_process['primal_value'],
                                        p_lpi_lp['primal_value'],
                                        d_lpi_lp['primal_value'],
                                        p_lpi['dual_value'],
                                        d_lpi['dual_value'],
                                        p_lpi_process['dual_value'],
                                        d_lpi_process['dual_value'],
                                        p_lpi_lp['dual_value'],
                                        d_lpi_lp['dual_value']
                                        ])), msg)

        check = lambda x: all([np.isclose(v, truth_x[k])
                               for k, v in x.items()])
        self.assertTrue(all(map(check, [p["x"], d["x"],
                                        p_lp["x"], d_lp["x"]])), msg)

        check = lambda x: all([np.isclose(v, truth_x_lpi[k])
                               for k, v in x.items()])
        self.assertTrue(all(map(check, [p_lpi["x"], p_lpi_process['x'],
                                        d_lpi["x"], d_lpi_process['x'],
                                        p_lpi_lp["x"], d_lpi_lp["x"]])), msg)

    def test_partially_known_objective(self):
        """Check that semiknown_moments are correctly processed."""
        problem = {
            "objective":  {'x': -1, 'y': -2, 'z': -1},
            "known_vars": {'1': 1, 'y': 25},
            "inequalities":  [{'x': 1, 'y': 1, 'z': 1, '1': -26},
                              {'x': 1, '1': -3},
                              {'y': 1, '1': -4},
                              {'z': 1, '1': -1}],
            "equalities": [{'x': -5, 'y': 1, 'z': -2, '1': -7}],
            "semiknown_vars": {'z': (0.5, 'x')}
        }
        problem_mat = convert_dicts(**problem, variables=var)
        p_lpi = solveSDP_MosekFUSION(**problem,
                                     solve_dual=False,
                                     process_constraints=False)
        p_lpi_process = solveSDP_MosekFUSION(**problem,
                                             solve_dual=False,
                                             process_constraints=True)
        d_lpi = solveSDP_MosekFUSION(**problem,
                                     solve_dual=True,
                                     process_constraints=False)
        d_lpi_process = solveSDP_MosekFUSION(**problem,
                                             solve_dual=True,
                                             process_constraints=True)
        p_lpi_lp_sparse = solveLP_sparse(**problem_mat,
                                         variables=var,
                                         solve_dual=False)
        d_lpi_lp_sparse = solveLP_sparse(**problem_mat,
                                         variables=var,
                                         solve_dual=True)
        truth_obj_lpi = -109/2
        truth_x_lpi = {'x': 3, 'z': 3/2}

        msg = "The dual and primal solutions are not equal when " + \
              "processing semiknown constraints and partially-known objective."


        check = lambda x: all([np.isclose(v, truth_x_lpi[k])
                               for k, v in truth_x_lpi.items()])
        self.assertTrue(check(p_lpi["x"]), msg)
        self.assertTrue(check(p_lpi_process["x"]), msg)
        self.assertTrue(check(p_lpi_lp_sparse["x"]), msg)
        self.assertTrue(check(d_lpi["x"]), msg)
        self.assertTrue(check(d_lpi_process["x"]), msg)
        self.assertTrue(check(d_lpi_lp_sparse["x"]), msg)

        check = lambda x: np.isclose(x, truth_obj_lpi)

        vals = [
            p_lpi['primal_value'],
            p_lpi['dual_value'],
            p_lpi_process['primal_value'],
            p_lpi_process['dual_value'],
            p_lpi_lp_sparse['primal_value'],
            p_lpi_lp_sparse['dual_value'],
            d_lpi_process['primal_value'],
            d_lpi_process['dual_value'],
            d_lpi['primal_value'],
            d_lpi['dual_value'],
            d_lpi_lp_sparse['primal_value'],
            d_lpi_lp_sparse['dual_value'],
        ]
        self.assertTrue(all(map(check, vals)), msg + "\n" +f"{vals} + vs {truth_obj_lpi}")


class TestTools(unittest.TestCase):
    def test_to_sparse(self):
        with self.subTest(msg="Convert dictionary to sparse matrix"):
            known_vars = {'y': 0, 'x': -2, 'z': 9}
            variables = ['x', 'y', 'z']
            expected_mat = coo_matrix(([-2, 0, 9], ([0, 0, 0], [0, 1, 2])),
                                      shape=(1, 3))
            actual_mat = to_sparse(known_vars, variables)
            self.assertEqual((expected_mat - actual_mat).nnz, 0,
                             "The dictionary was not correctly converted to a "
                             "sparse matrix.")
        with self.subTest(msg="Convert list of dictionaries to sparse matrix"):
            inequalities = simple_lp["inequalities"]
            variables = ['1', 'w', 'x', 'y', 'z']
            expected_mat = coo_matrix(([2, -1, 5, -1, 1/2, -1, 1, 1],
                                       ([0, 0, 1, 1, 2, 2, 3, 3],
                                       [0, 2, 0, 3, 0, 4, 0, 1])),
                                      shape=(4, 5))
            actual_mat = to_sparse(inequalities, variables)
            self.assertEqual((expected_mat - actual_mat).nnz, 0,
                             "The list of dictionaries was not correctly "
                             "converted to a sparse matrix.")

    def test_convert_dicts(self):
        dict_lp = deepcopy(simple_lp)
        semiknown = dict_lp["semiknown_vars"] = {'x': (4, 'y'), 'w': (-2, 'z')}
        dict_lp["lower_bounds"] = {'x': 0, 'y': -2, 'z': 3}
        dict_lp["upper_bounds"] = {'w': 5, 'x': 0, 'y': -1}
        mat_lp = {k: to_sparse(v, var) for k, v in dict_lp.items()
                  if k != "semiknown_vars"}
        semiknown_mat = coo_matrix(([1, -4, 1, 2],
                                    ([0, 0, 1, 1], [2, 3, 1, 4])),
                                   shape=(len(semiknown), len(var)))
        mat_lp["equalities"] = vstack((mat_lp["equalities"], semiknown_mat))
        mixed_lp = {k: dict_lp[k] for k in ("objective", "semiknown_vars")}
        mixed_lp.update({k: mat_lp[k]
                         for k in ("known_vars", "upper_bounds")})

        with self.subTest(msg="Variables not passed"):
            with self.assertRaises(AssertionError):
                convert_dicts(**dict_lp)
        with self.subTest(msg="All arguments are matrices"):
            expected_args = {}
            actual_args = convert_dicts(**mat_lp, variables=var)
            self.assertEqual(expected_args, actual_args)

        types = {"All dictionaries": (mat_lp, dict_lp),
                 "Mixed type": ({"objective": mat_lp["objective"],
                                 "equalities": semiknown_mat},
                                mixed_lp)}
        for type_name, (expected_args, to_convert) in types.items():
            with self.subTest(msg="Convert arguments of type:", i=type_name):
                actual_args = convert_dicts(**to_convert, variables=var)
                self.assertEqual(len(expected_args), len(actual_args),
                                 "The number of returned arguments is not "
                                 "correct.")
                for arg in expected_args:
                    (exp_arg, act_arg) = (expected_args[arg], actual_args[arg])
                    self.assertEqual((exp_arg - act_arg).nnz, 0,
                                     f"{arg} is not equal: "
                                     f"{exp_arg.toarray()} != "
                                     f"{act_arg.toarray()}.")



"""
This file contains helper functions to optimize (functions of) symbolic
variables under the constraint that the InflationLP or InflationSDP instance
they implicitly define must be feasible. Useful for exploring the set of
parametrically-defined distribution for which quantum inflation (at specific
hierarchy levels) is

@authors: Emanuel-Cristian Boghiu, Elie Wolfe, Alejandro Pozas-Kerstjens
"""
import numpy as np
import sympy as sp

from numbers import Real
from scipy.optimize import bisect, minimize, Bounds
from sympy.utilities.lambdify import lambdify
from typing import Dict, Tuple, Union
from functools import lru_cache

from . import InflationLP, InflationSDP
from .utils import make_numerical
from .sdp.monomial_classes import CompoundMomentSDP


def max_within_feasible(program: Union[InflationLP, InflationSDP],
                        symbolic_values: Dict[CompoundMomentSDP,
                                              sp.core.expr.Expr],
                        method: str,
                        return_last_certificate=False,
                        **kwargs) -> Union[float,
                                           Tuple[float, dict]]:
    """Maximize a single real variable within the set of feasible moment
    matrices determined by an ``InflationSDP``. The dependence of the moment
    matrices in the variable is specified by an assignment of monomials in the
    moment matrix to arbitrary expressions of the variable. This is useful for
    finding (bounds for) critical visibilities of distributions beyond which
    they are impossible to generate in a given quantum causal scenario.

    Parameters
    ----------
    program : Union[InflationLP, InflationSDP]
        The problem under which to carry the optimization.
    symbolic_values : Dict[CompoundMomentSDP, Callable]
        The correspondence between monomials in the SDP problem and symbolic
        expressions depending on the variable to be optimized.
    method : str
        Technique used for optimization. Currently supported: ``"bisection"``
        for bisection algorithms, and ``"dual"`` for exploitation of the
        certificates of infeasibility (typically much fewer iteration steps).
    return_last_certificate : bool, optional
        Whether to return, along with the maximum value of the parameter, a
        separating surface that leaves the set of positive-semidefinite moment
        matrices in its positive side and evaluates to 0 in the maximum value
        reported.

    **kwargs
        Instructions on which extra symbolic values to assign numbers and
        options to be passed to the optimization routines (bounds, precision,
        tolerance, ...).

    Returns
    -------
    float
        The maximum value that the parameter can take under the set of
        distributions compatible with an inflation (for LPs) or under the set of
        positive-semidefinite moment matrices (for SDPs). This is the output
        when ``return_last_certificate=False``.
    Tuple[float, dict]
        The maximum value that the parameter can take under the set of
        distributions compatible with an inflation (for LPs) or under the set of
        positive-semidefinite moment matrices (for SDPs), and a corresponding
        separating surface (a root of the function corresponds to the critical
        feasible value of the parameter reported). This is the output when
        ``return_last_certificate=True``.
    """
    assert method in ["bisection", "dual"], \
        "Unknown optimization method. Please use \"bisection\" or \"dual\"."
    variables = set()
    for expr in symbolic_values.values():
        if not isinstance(expr, Real):
            variables.update(expr.atoms(sp.Symbol))
    assert len(variables) != 0, \
        "No variable to be optimized detected"
    assert len(variables) == 1, \
        "Only optimization of a single variable is supported"
    param = variables.pop()
    kwargs.update({"return_last_certificate": return_last_certificate})

    if method == "bisection":
        return _maximize_via_bisect(program, symbolic_values, param, **kwargs)
    elif method == "dual":
        return _maximize_via_dual(program, symbolic_values, param, **kwargs)


###############################################################################
# OPTIMIZATION METHODS                                                        #
###############################################################################
def _maximize_via_bisect(program: Union[InflationLP, InflationSDP],
                         symbolic_values: Dict[CompoundMomentSDP,
                                               sp.core.expr.Expr],
                         param: sp.core.symbol.Symbol,
                         **kwargs) -> Union[float,
                                            Tuple[float, dict]]:
    """Implement the maximization of a variable within the feasible set of
    distributions using SciPy's bisection algorithm.


    Parameters
    ----------
    program : Union[InflationLP, InflationSDP]
        The problem under which to carry the optimization.
    symbolic_values : Dict[CompoundMomentSDP, Callable]
        The correspondence between monomials in the problem and symbolic
        expressions depending on the variable to be optimized.
    param : sympy.core.symbol.Symbol
        The variable to be optimized.

    **kwargs
        Additional arguments to ``program.set_values()`` and
        ``scipy.optimize.bisect()``.

    Returns
    -------
    float
        The maximum value that the parameter can take under the set of
        distributions compatible with an inflation (for LPs) or under the set of
        positive-semidefinite moment matrices (for SDPs). This is the output
        when ``return_last_certificate=False``.
    Tuple[float, sympy.core.add.Add]
        The maximum value that the parameter can take under the set of
        distributions compatible with an inflation (for LPs) or under the set of
        positive-semidefinite moment matrices (for SDPs), and a corresponding
        separating surface (a root of the function corresponds to the critical
        feasible value of the parameter reported). This is the output when
        ``return_last_certificate=True``.
    """
    bounds         = kwargs.get("bounds", np.array([0.0, 1.0]))
    only_specified = kwargs.get("only_specified_values", False)
    return_last    = kwargs.get("return_last_certificate", False)
    precision      = kwargs.get("precision", 1e-4)
    use_lpi        = kwargs.get("use_lpi_constraints", False)
    verbose        = kwargs.get("verbose", False)
    solve_dual     = kwargs.get("solve_dual", True)
    solverparameters = kwargs.get("solverparameters", None)
    # Prepare bisect kwargs
    bisect_kwargs = {}
    for kwarg in ["args", "rtol", "maxiter", "full_output", "disp"]:
        try:
            bisect_kwargs[kwarg] = kwargs[kwarg]
        except KeyError:
            pass
    try:
        bisect_kwargs["xtol"] = kwargs["xtol"]
    except KeyError:
        bisect_kwargs["xtol"] = precision

    discovered_certificates = []

    @lru_cache(maxsize=2, typed=False)
    def f(value: float) -> float:
        evaluated_values = make_numerical(symbolic_values, {param: value})
        program.set_values(evaluated_values,
                       use_lpi_constraints=use_lpi,
                       only_specified_values=only_specified)
        program.solve(feas_as_optim=True,
                  solve_dual=solve_dual,
                  solverparameters=solverparameters)
        if verbose:
            print(f"Parameter = {value:<6.4g}   " +
                  f"Maximum smallest eigenvalue: {program.objective_value:10.4g}")
        discovered_certificate = program.certificate_as_dict()
        if len(discovered_certificate):  # Checking if certificate has any nonzero coefficients
            discovered_certificates.append((value, discovered_certificate))
        return program.objective_value+precision/2048
    assert f(bounds[1])+precision/2048 < 0, f"Certificate of infeasibility for v=1 is {f(bounds[1])}, but must be <{precision/2048}."
    assert f(bounds[0])+precision/2048 > 0, f"Certificate of feasibility for v=0 is {f(bounds[0])}, but must be nonnegative."
    crit_param = bisect(f, bounds[0], bounds[1], **bisect_kwargs, full_output=False)
    if return_last:
        return crit_param, discovered_certificates[-1][-1]
    else:
        return crit_param


def _maximize_via_dual(program: Union[InflationLP, InflationSDP],
                       symbolic_values: Dict[CompoundMomentSDP,
                                             sp.core.expr.Expr],
                       param: sp.core.symbol.Symbol,
                       **kwargs) -> Union[float,
                                          Tuple[float, dict]]:
    """Implement the maximization of a variable within the feasible set of
    distributions exploiting the certificates of infeasibility. For a given
    value of the parameter, a separating surface that leaves the set of
    fesible distributions in its positive side is extracted. The next value of
    the parameter used is that which evaluates the surface to 0.

    Parameters
    ----------
    program : Union[InflationLP, InflationSDP]
        The problem under which to carry the optimization.
    symbolic_values : Dict[CompoundMomentSDP, Callable]
        The correspondence between monomials in the problem and symbolic
        expressions depending on the variable to be optimized.
    param : sympy.core.symbol.Symbol
        The variable to be optimized.

    **kwargs
        Additional arguments to ``program.set_values()``, bounds of the
        optimization interval, precision of the optimization, verbosity and
        whether returning the last computed separating surface.

    Returns
    -------
    float
        The maximum value that the parameter can take under the set of
        distributions compatible with an inflation (for LPs) or under the set of
        positive-semidefinite moment matrices (for SDPs). This is the output
        when ``return_last_certificate=False``.
    Tuple[float, dict]
        The maximum value that the parameter can take under the set of
        distributions compatible with an inflation (for LPs) or under the set of
        positive-semidefinite moment matrices (for SDPs), and a corresponding
        separating surface (a root of the function corresponds to the critical
        feasible value of the parameter reported). This is the output when
        ``return_last_certificate=True``.
    """
    bounds         = kwargs.get("bounds", np.array([0.0, 1.0]))
    only_specified = kwargs.get("only_specified_values", False)
    precision      = kwargs.get("precision", 1e-4)
    return_last    = kwargs.get("return_last_certificate", False)
    use_lpi        = kwargs.get("use_lpi_constraints", False)
    verbose        = kwargs.get("verbose", False)
    solve_dual     = kwargs.get("solve_dual", True)
    solverparameters = kwargs.get("solverparameters", None)

    symbol_names = {str(key): val for key, val in symbolic_values.items()}
    func_to_minimize = lambdify([param], -param, modules='numpy', cse=True)
    sp_bounds = Bounds(lb=bounds[0], ub=bounds[1])
    x0 = np.array([bounds[0]])
    new_ub = bounds[1]-precision
    old_ub = np.inf
    discovered_certificates = []
    # Get a certificate for a value, find the critical value for it (that which
    # makes the certificate zero), and repeat with this new value
    while new_ub+precision < old_ub:
        old_ub = new_ub
        evaluated_values = make_numerical(symbolic_values,
                                          {param: new_ub+precision})
        program.set_values(evaluated_values,
                           use_lpi_constraints=use_lpi,
                           only_specified_values=only_specified)
        program.solve(feas_as_optim=True,
                      solve_dual=solve_dual,
                      solverparameters=solverparameters)

        # The feasible region is where the certificate is positive
        nonneg_expr = sum(symbol_names[var] * coeff
                          for var, coeff in
                          program.solution_object["dual_certificate"].items())
        nonneg_func = lambdify([param],
                               nonneg_expr,
                               modules="numpy",
                               cse=True)
        constraints = {"type": "ineq", "fun": nonneg_func}
        solution = minimize(func_to_minimize, x0,
                            method='SLSQP',
                            bounds=sp_bounds,
                            constraints=constraints,
                            options={'disp': False})
        new_ub = solution['x'][0]
        discovered_certificates.append((new_ub , program.certificate_as_dict()))
        if verbose:
            print(f"Current critical value: {new_ub} (seeded by {old_ub+precision})")
    assert len(discovered_certificates)>1, """
    Critical error - optimization failed to yield a useful certificate
    even when given the initial (infeasible) seed.
    """

    crit_param = max(new_ub, old_ub)
    if return_last:
        return crit_param, discovered_certificates[-1][-1]
    else:
        return crit_param



"""
This file contains helper functions to manipulate monomials and generate moment
matrices. The functions in this file can be accelerated by JIT compilation in
numba.

@authors: Emanuel-Cristian Boghiu, Elie Wolfe, Alejandro Pozas-Kerstjens
"""
import numpy as np

from typing import List

try:
    from numba import jit
    from numba.types import bool_, void
    from numba.types import uint8 as uint8_
    from numba.types import intc as int_
    nopython = True
except ImportError:
    print("Warning: NOT using Numba for JIT compilation.")
    def jit(*args, **kwargs):
        return lambda f: f
    bool_    = bool
    nopython = False
    uint8_   = np.uint8
    int_     = np.intc
    void     = None

cache    = True
if not nopython:
    from scipy.sparse.csgraph import connected_components
    bool_  = bool
    uint8_ = np.uint8


###############################################################################
# ABSTRACT OPERATIONS ON MONOMIALS                                            #
###############################################################################
@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def dot_mon(mon1: np.ndarray,
            mon2: np.ndarray) -> np.ndarray:
    """Returns ((mon1)^dagger)*mon2.

    For hermitian operators this is the same as reversed(mon1)*mon2. Since all
    parties commute, the output is ordered by parties. We do not assume any
    other commutation rules.

    Parameters
    ----------
    mon1 : numpy.ndarray
        Input monomial in 2d array format.
    mon2 : numpy.ndarray
        Input monomial in 2d array format.

    Returns
    -------
    numpy.ndarray
        The monomial ordered by parties.

    Examples
    --------
    >>> mon1 = np.array([[1,2,3],[4,5,6]])
    >>> mon2 = np.array([[7,8,9]])
    >>> dot_mon(mon1, mon2)
    np.array([[4,5,6],[1,2,3],[7,8,9]])
    """
    return np.concatenate((reverse_mon(mon1), mon2))


@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def mon_lexsorted(mon: np.ndarray,
                  lexorder: np.ndarray) -> np.ndarray:
    """Sorts a monomial lexicographically.

    Parameters
    ----------
    mon : numpy.ndarray
        Input monomial in 2d array format.
    lexorder : numpy.ndarray
        Matrix with rows as operators where the index of the row gives
        the lexicographic order of the operator.

    Returns
    -------
    numpy.ndarray
        Sorted monomial.
    """
    mon_lexrepr = nb_mon_to_lexrepr(mon, lexorder)
    return mon[np.argsort(mon_lexrepr, kind="quicksort")]


@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def nb_classify_disconnected_components(adj_mat: np.ndarray) -> np.ndarray:
    """Given a boolean matrix where each cell indicates whether the supports of
    the operator denoting the row and the operator denoting the column overlap,
    generate a list determining to which disconnected component each operator
    belongs to.

    Parameters
    ----------
    adj_mat : numpy.ndarray
        Boolean 2d array where each cell indicates whether the supports of the
        operator denoting the row and the operator denoting the column overlap.

    Returns
    -------
    numpy.ndarray
        A list of integers of size the number of operators used for creating
        adj_mat, where each integer indexes the disconnected component the
        corresponding operator belongs to.
    """
    if not nopython:
        return connected_components(adj_mat,
                                    directed=False,
                                    return_labels=True)[-1]
    # See https://stackoverflow.com/a/9112588 for inspiration of the method
    n = len(adj_mat)
    if n <= 1 or adj_mat.all():
        return np.zeros((n,), dtype=uint8_)
    component_labels = np.zeros((n,), dtype=uint8_)
    component_counter = 1
    for i in range(n):
        if not component_labels[i]:
            old_component = np.logical_not(np.ones((n,), dtype=uint8_))
            new_component = old_component.copy()
            new_component[i] = True
            search_next = np.logical_xor(new_component, old_component)
            while search_next.any():
                old_component = new_component.copy()
                new_component = np.logical_or(
                    new_component,
                    adj_mat[search_next].sum(axis=0).astype(bool_))
                search_next = np.logical_xor(new_component, old_component)
            component_labels[new_component] = component_counter
            component_counter += 1
    return (component_labels-1).astype(np.uint8)


@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def nb_is_knowable(monomial: np.ndarray) -> bool_:
    """Determine whether a given atomic monomial admits an identification with
    a probability of the original scenario.

    Parameters
    ----------
    monomial : np.ndarray
        List of operators, denoted each by a list of indices

    Returns
    -------
    bool
        Whether the monomial is knowable or not.
    """
    if len(monomial) <= 1:
        return True
    # Knowable monomials have at most one operator per party and one copy of
    # each source in the DAG
    parties = monomial[:, 0]
    if len(np.unique(parties)) != len(monomial):
        return False
    for source in monomial.T[1:-2]:
        if len(np.unique(source[np.flatnonzero(source)])) > 1:
            return False
    return True


@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def nb_lexorder_idx(operator: np.ndarray,
                    lexorder: np.ndarray) -> int:
    """Return the unique integer corresponding to the lexicographic ordering of
    the operator. If ``operator`` is not found in ``lexorder``, no value is
    returned.

    Parameters
    ----------
    operator : numpy.ndarray
        Array of integers representing the operator.
    lexorder : numpy.ndarray
        Matrix with rows as operators where the index of the row gives
        the lexicographic ordering.

    Returns
    -------
    int
        The index of the operator in the lexorder matrix.
    """
    for i in range(lexorder.shape[0]):
        if np.array_equal(lexorder[i, :], operator):
            return i
    return -1

@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def nb_mon_to_lexrepr(mon: np.ndarray,
                      lexorder: np.ndarray) -> np.array:
    """Convert a monomial to its lexicographic representation, in the form of
    an array of integers representing the lexicographic rank of each operator.

    Parameters
    ----------
    mon : numpy.ndarray
        Input monomial in 2d array format.
    lexorder : numpy.ndarray
        Matrix with rows as operators where the index of the row gives
        the lexicographic order of the operator.

    Returns
    -------
    np.array
        Monomial as array of integers, where each integer is the hash
        of the corresponding operator.
    """
    lex = np.empty(mon.shape[0], dtype=int_)
    for i in range(mon.shape[0]):
        lex[i] = nb_lexorder_idx(mon[i], lexorder)
    return lex


@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def nb_monomial_to_components(monomial: np.ndarray) -> np.ndarray:
    """Wrapper for obtaining the list of disconnected components of a monomial.

    Parameters
    ----------
    monomial : numpy.ndarray
        Monomial in 2d array form.

    Returns
    -------
    numpy.ndarray
        A vector where each integer gives the component associated with the
        operator of that index.

    Examples
    --------
    >>> monomial = np.array([[1, 0, 1, 1, 0, 0],
                             [2, 1, 0, 2, 0, 0],
                             [1, 0, 3, 3, 0, 0],
                             [3, 3, 5, 0, 0, 0],
                             [3, 1, 4, 0, 0, 0],
                             [3, 6, 6, 0, 0, 0],
                             [3, 4, 5, 0, 0, 0]])
    >>> factorised = nb_monomial_to_components(monomial)
    [0, 1, 2, 3, 1, 4, 3]
    """
    n = len(monomial)
    if n <= 1:
        return np.zeros(n, dtype=uint8_)
    return nb_classify_disconnected_components(nb_overlap_matrix(
        monomial[:, 1:-2]))


@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def nb_overlap_matrix(inflation_indxs: np.ndarray) -> np.ndarray:
    """Given a list of inflation indices for a number of operators, generate
    a boolean matrix whose entries denote whether the supports of the operator
    indexed by the row and of the operator indexed by the column overlap.

    Parameters
    ----------
    inflation_indxs : numpy.ndarray
        A list of inflation indices for a collection of operators.

    Returns
    -------
    numpy.ndarray
        The "adjacency matrix" whose entries denote whether the supports of the
        operator indexed by the row and of the operator indexed by the column
        overlap.
    """
    n = len(inflation_indxs)
    adj_mat = np.eye(n, dtype=bool_)
    for i in range(1, n):
        inf_indices_i = inflation_indxs[i]
        for j in range(i):
            inf_indices_j = inflation_indxs[j]
            if nb_exists_shared_source(inf_indices_i, inf_indices_j):
                adj_mat[i, j] = True
    adj_mat = np.logical_or(adj_mat, adj_mat.T)
    return adj_mat


@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def nb_indices_of_more_than_one_op_per_party_per_factor(monomial: np.ndarray,
                                                        sandwich_positivity=True) -> np.ndarray:
    r"""If sandwich_positivity=True, this removes sandwiching/pinching from the
    monomial. This is, it converts the monomial represented by
    :math:`U A U^\dagger` into :math:`A`.
    Regardless, it then factorises the monomial into components, yielding
    a boolean vector where True indicates an operator which appears in the same
    factor as another of the same party.

    Parameters
    ----------
    monomial : numpy.ndarray
        Input monomial.
    sandwich_positivity : bool, optional
        Whether to consider sandwiching. By default ``True``.

    Returns
    -------
    numpy.ndarray
        The monomial without sandwiches.
    """
    picklist = np.logical_not(np.ones(len(monomial)))
    parties = np.unique(monomial[:, 0])
    for party in parties:
        indices_for_this_party = np.flatnonzero(monomial[:, 0] == party)
        party_monomial = monomial[indices_for_this_party]
        party_monomial_comp_labels = nb_monomial_to_components(party_monomial)
        for i in range(party_monomial_comp_labels.max() + 1):
            indices_for_this_factor = np.flatnonzero(
                party_monomial_comp_labels == i)
            factor = party_monomial[indices_for_this_factor]
            # We now only return factors with more than one operator for a single party!
            if sandwich_positivity:
                while len(factor) and np.array_equal(factor[0], factor[-1]):
                    indices_for_this_factor = indices_for_this_factor[1:-1]
                    factor = factor[1:-1]
            if len(factor) > 1:
                picklist[indices_for_this_party[indices_for_this_factor]] = True
    return picklist

@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def nb_is_physical(monomial_in: np.ndarray, sandwich_positivity=True) -> bool_:
    r"""Determines whether a monomial is physical, this is, if it always has a
    non-negative expectation value.

    This code also supports the detection of "sandwiches", i.e., monomials
    of the form :math:`\langle \psi | A_1 A_2 A_1 | \psi \rangle` where
    :math:`A_1` and :math:`A_2` do not commute. In principle we do not know the
    value of this term. However, note that :math:`A_1` can be absorbed into
    :math:`| \psi \rangle` forming an unnormalised quantum state
    :math:`| \psi' \rangle`, thus :math:`\langle\psi'|A_2|\psi'\rangle`.
    Note that while we know the value :math:`\langle\psi |A_2| \psi\rangle`,
    we do not know :math:`\langle \psi' | A_2 | \psi' \rangle` because of
    the unknown normalisation, however we know it must be non-negative,
    :math:`\langle \psi | A_1 A_2 A_1 | \psi \rangle \geq 0`.
    This simple example can be extended to various layers of sandwiching.

    Parameters
    ----------
    monomial_in : numpy.ndarray
        Input monomial in 2d array format.
    sandwich_positivity : bool, optional
        Whether to consider sandwiching. By default ``True``.

    Returns
    -------
    bool
        Whether the monomial has always non-negative expectation or not.
    """
    return not nb_indices_of_more_than_one_op_per_party_per_factor(
        monomial_in, sandwich_positivity=sandwich_positivity).any()

@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def remove_projector_squares(mon: np.ndarray) -> np.ndarray:
    """Simplify the monomial by removing operator powers. This is because we
    assume projectors, P**2=P. This corresponds to removing duplicates of rows
    which are adjacent.

    Parameters
    ----------
    mon : numpy.ndarray
        Input monomial in 2d array format.

    Returns
    -------
    bool
        Monomial with powers of operators removed.
    """
    to_keep = np.ones(mon.shape[0], dtype=bool_)
    for i in range(1, mon.shape[0]):
        if np.array_equal(mon[i], mon[i - 1]):
            to_keep[i] = False
    return mon[to_keep]


@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def reverse_mon(mon: np.ndarray) -> np.ndarray:
    """Return the reversed monomial.

    This represents the Hermitian conjugate of the monomial, assuming that
    each operator is Hermitian.

    Parameters
    ----------
    mon : numpy.ndarray
        Input monomial in 2d array format.

    Returns
    -------
    numpy.ndarray
        Reversed monomial.
    """
    return np.flipud(mon)


@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def to_name(monomial: np.ndarray,
            names: List[str]) -> str:
    """Convert the 2d array representation of a monomial to a string.

    Parameters
    ----------
    monomial : numpy.ndarray
        Monomial as a matrix with each row being an integer array representing
        an operators.
    names : List[str]
        List of party names.

    Returns
    -------
    str
        The string representation of the monomial.

    Examples
    --------
    >>> to_name([[1 1,0,3], [4,1,2,6], [2,3,3,4]], ['a','bb','x','Z'])
    'a_1_0_3*Z_1_2_6*bb_3_3_4'
    """
    if len(monomial) == 0:
        return "1"

    monomial = monomial.tolist()
    return "*".join(["_".join([names[letter[0] - 1]]
                              + [str(i) for i in letter[1:]])
                     for letter in monomial])


###############################################################################
# OPERATIONS ON MONOMIALS RELATED TO INFLATION                                #
###############################################################################
@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def apply_source_perm(monomial: np.ndarray,
                      source: int,
                      permutation: np.ndarray) -> np.ndarray:
    """Apply a source swap to a monomial.

    Parameters
    ----------
    monomial : numpy.ndarray
        Input monomial in 2d array format.
    source : int
        The source that is being swapped.
    permutation : numpy.ndarray
        The permutation of the copies of the specified source. The format for
        the permutation here is to use indexing starting at one, so the
        permutation must be padded with a leading zero. The function
        ``inflation.sdp.quantum_tools.format_permutations`` converts
        a list of permutations to the necessary format.

    Returns
    -------
    numpy.ndarray
        Input monomial with the specified source swapped.
    """
    new_factors = monomial.copy()
    new_factors[:, 1 + source] = permutation[new_factors[:, 1 + source]]
    return new_factors


@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def commutation_matrix(lexorder: np.ndarray,
                       sources_to_check_for_pairwise: np.ndarray,
                       commuting=False) -> np.ndarray:
    """Build a matrix encoding of which operators commute according to the
    function ``nb_operators_commute``. Rows and columns are indexed by
    operators, and each element is a 0 if the operators in the row and in the
    column commute or 1 if they do not.

    Parameters
    ----------
    lexorder : numpy.ndarray
        Matrix with rows as operators where the index of the row gives the
        lexicographic order of the operator.
    quantum_sources: numpy.ndarray
        List of integers denoting the columns that in the 2D array encoding
        corresponding to inflation indices of quantum sources (as opposed to
        classical sources). Example: np.array([1, 3]) implies that the 1st
        and 3rd columns of an operator in 2D array form correspond to inflation
        indices that act on quantum sources.
    sources_to_check_for_pairwise: numpy.ndarray
        A 3-index tensor boolean array which keep track of which pairs of parties
        have which quantum sources in common via intermediate quantum latents.
    commuting : bool
        Whether all the monomials commute. In such a case, the trivial all-zero
        array is returned.

    Returns
    -------
    numpy.ndarray
        Matrix whose entry :math:`(i,j)` has value 1 if the operators with
        lexicographic ordering :math:`i` and :math:`j` do not commute, and
        value 0 if they commute.
    """
    notcomm = np.zeros((lexorder.shape[0], lexorder.shape[0]), dtype=bool_)
    if not commuting:
        for i in range(lexorder.shape[0]):
            for j in range(i + 1, lexorder.shape[0]):
                notcomm[i, j] = not nb_operators_commute(lexorder[i],
                                                         lexorder[j],
                                                         sources_to_check_for_pairwise)
        notcomm = notcomm + notcomm.T
    return notcomm


@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def nb_all_commuting_q(mon: np.ndarray,
                       lexorder: np.ndarray,
                       notcomm: np.ndarray) -> bool_:
    """Check if all operators in ``mon`` commute.

    Parameters
    ----------
    mon : numpy.ndarray
        Input monomial in 2d array format.
    lexorder : numpy.ndarray
        A matrix where each row is an operator, and the `i`-th row stores the
        operator with lexicographic rank `i`.
    notcomm : numpy.ndarray
        Matrix of commutation relations. Each operator can be identified by an
        integer `i` which also doubles as its lexicographic rank. Given two
        operators with ranks `i`, `j`, ``notcomm[i, j]`` is 1 if the operators
        do not commute, and 0 if they do.

    Returns
    -------
    bool
        Return ``True`` if all operators commute, and ``False`` otherwise.
    """
    if len(mon) <= 1:
        return True
    lexmon = nb_mon_to_lexrepr(mon, lexorder)
    sub_notcomm = notcomm[lexmon, :][:, lexmon]
    return not sub_notcomm.any()


@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def nb_exists_shared_source(inf_indices1: np.ndarray,
                            inf_indices2: np.ndarray) -> bool_:
    """Determine if there is any common source referred to by two
     specifications of inflation (a.k.a. 'copy') indices.

    Parameters
    ----------
    inf_indices1 : numpy.ndarray
        An array of integers indicating which copy index of each source is
        being referenced.
    inf_indices2 : numpy.ndarray
        An array of integers indicating which copy index of each source is
        being referenced.

    Returns
    -------
    bool
        ``True`` if ``inf_indices1`` and ``inf_indices2`` share a source in
        common, ``False`` if they do not.
    """
    common_sources = np.logical_and(inf_indices1, inf_indices2)
    if not np.any(common_sources):
        return False
    return not np.subtract(inf_indices1[common_sources],
                           inf_indices2[common_sources]).all()
@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def nb_lexmon_to_canonical(lexmon: np.ndarray,
                           notcomm: np.ndarray) -> np.ndarray:
    """Brings a monomial, input as the indices of the operators in the
    lexicographic ordering, to canonical form with respect to commutations.

    Parameters
    ----------
    lexmon : numpy.ndarray
        Monomial as an array of integers, where each integer represents
        the lexicographic order of an operator.
    notcomm : numpy.ndarray
        Matrix of commutation relations, given in the format specified by
        `inflation.sdp.fast_npa.commutation_matrix`.

    Returns
    -------
    numpy.ndarray
        Monomial in canonical form with respect to commutations.
    """
    result = np.empty((0,), dtype=int_)
    leftover = lexmon
    while leftover.shape[0] > 1:
        lexmon = leftover
        ########################################################################
        # A subroutine that splits lexmon into two monomials, m1 and m2,
        # such that m1 is appended to result, and m2 is the new lexmon for a
        # new iteration of the while loop.
        # This subroutine should be defined in a function but numba
        # gives problems with returning tuples (m1,m2) from functions.
        if lexmon.shape[0] <= 1:
            m1, m2 = lexmon, np.empty((0,), dtype=int_)
        else:
            # Take only the rows and columns of notcomm that appear in the
            # monomial, in the correct order.
            sub_notcomm = notcomm[lexmon, :][:, lexmon]
            if sub_notcomm.all():
                m1, m2 = lexmon, np.empty((0,), dtype=int_)
            elif not sub_notcomm.any():
                m1, m2 = np.sort(lexmon), np.empty((0,), dtype=int_)
            else:
                minimum = lexmon[0]
                minimum_idx = 0
                for op in range(1, lexmon.shape[0]):
                    # Find the lowest position where we can move op
                    nc_ops_position = np.where(sub_notcomm[op, :op] == 1)[0]
                    if nc_ops_position.size < 1:
                        if lexmon[op] < minimum:
                            minimum_idx = op
                            minimum     = lexmon[op]
                if minimum <= lexmon[0]:
                    m1 = np.array([lexmon[minimum_idx]])
                    m2 = np.concatenate((lexmon[:minimum_idx],
                                        lexmon[minimum_idx + 1:]))
                else:
                    m1 = np.array([lexmon[0]])
                    m2 = lexmon[1:]
        ########################################################################
        result = np.concatenate((result, m1))
        leftover = m2
    # Depending on how the while loop ends, we may have a leftover monomial
    if leftover.shape[0] != 0:
        result = np.concatenate((result, leftover))
    return result

@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def nb_operators_commute(operator1: np.ndarray,
                         operator2: np.ndarray,
                         sources_to_check_for_pairwise: np.ndarray
                         ) -> bool_:
    """Determine if two operators commute. Currently, this only takes into
    account commutation coming from inflation and settings.

    Parameters
    ----------
    operator1 : numpy.ndarray
        Operator as an array of integers.
    operator2 : numpy.ndarray
        Operator as an array of integers.
    sources_to_check_for_pairwise: numpy.ndarray
        A 3-index tensor boolean array which keep track of which pairs of parties
        have which quantum or classical sources in common via intermediate latents.

    Returns
    -------
    bool
        ``True`` if ``operator1`` and ``operator2`` commute, ``False`` if they
        do not.

    Examples
    --------
    A^11_00 does not commute with A^12_00 because they overlap on source 1.
    >>> nb_operators_commute(np.array([1, 1, 1, 0, 0]),
                             np.array([1, 1, 2, 0, 0]),  np.array([[[2,2],[0,0]],[[0,0],[2,2]]))
    False
    
    A^11_00 commutes with A^12_00 because source 1 is classical.
    >>> nb_operators_commute(np.array([1, 1, 1, 0, 0]),
                             np.array([1, 1, 2, 0, 0]),  np.array([[[1,2],[0,0]],[[0,0],[1,2]]))
    True
    """
    # Case 0: Different parties commute
    party1 = operator1[0] - 1
    party2 = operator2[0] - 1
    relevant_sources = sources_to_check_for_pairwise[party1, party2]
    common_source_positions_all = np.greater_equal(relevant_sources, 1)
    common_source_positions_quantum = np.greater_equal(relevant_sources, 2)

    # Case 1: no common quantum source TYPE
    if not common_source_positions_quantum.any():
        return True
    # Case 2: yes common quantum source TYPE, but different INDICES across all quantum source
    sources1 = operator1[1:-2]
    sources2 = operator2[1:-2]
    if np.subtract(sources1[common_source_positions_quantum],
                   sources2[common_source_positions_quantum]).all():
        return True
    # Case 3: overlapping quantum source type and index, but not all sources match indices
    if not np.array_equal(sources1[common_source_positions_all],
                          sources2[common_source_positions_all]):
        return False
    # Case 4: All sources of type common to both operators match indices
    # # Case 4a: different parties
    if party1 != party2:
        return True
    # Case 4b: Same party (check the settings)
    return operator1[-2] == operator2[-2]

# Node: cache is set to False because of a bug in numba, which affects
# nb_lexmon_to_canonical which this function uses
@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def to_canonical_1d_internal(lexmon: np.ndarray,
                             notcomm: np.ndarray,
                             orthomat: np.ndarray,
                             commuting=False,
                             apply_only_commutations=False) -> np.ndarray:
    """Brings a monomial to canonical form with respect to commutations,
    removing square projectors, and identifying orthogonality.

    Parameters
    ----------
    lexmon : numpy.ndarray
        Input monomial in 1D array format.
    notcomm : numpy.ndarray
        Matrix of commutation relations. Each operator can be identified by an
        integer `i` which also doubles as its lexicographic rank. Given two
        operators with ranks `i`, `j`, ``notcomm[i, j]`` is 1 if the operators
        do not commute, and 0 if they do.
    orthomat : numpy.ndarray
        Matrix of orthogonality relations. Each operator can be identified by an
        integer `i` which also doubles as its lexicographic rank. Given two
        operators with ranks `i`, `j`, ``notcomm[i, j]`` is 1 if the operators
        are orthogonal, and 0 if they do.
    commuting : bool, optional
        Whether all the variables in the problem commute or not. By default
        ``False``.
    apply_only_commutations : bool, optional
        If ``True``, skip the removal of projector squares and the test to see
        if the monomial is equal to zero. By default ``False``.

    Returns
    -------
    numpy.ndarray
        The monomial in canonical form with respect to some commutation
        relationships.
    """
    if len(lexmon) <= 1:
        return lexmon
    else:
        if commuting:
            newlexmon = np.sort(lexmon)
        else:
            newlexmon = nb_lexmon_to_canonical(lexmon, notcomm)
    if apply_only_commutations:
        return newlexmon
    else:
        if lexmon_is_zero(newlexmon, orthomat):
            newlexmon = np.zeros((1,), dtype=int_)
        return remove_projector_squares(newlexmon)


@jit(nopython=nopython, cache=cache, forceobj=not nopython)
def lexmon_is_zero(lexmon: np.ndarray, orthomat: np.ndarray) -> bool_:
    if np.array_equal(lexmon, [0]):
        return True
    for i in range(1, lexmon.shape[0]):
        if orthomat[lexmon[i], lexmon[i-1]]:
            return True
    return False



"""
This file contains classes for defining probabilities of events or expectation
values of monomials in convex programming relaxations of inflation.

@authors: Emanuel-Cristian Boghiu, Elie Wolfe, Alejandro Pozas-Kerstjens
"""
import numpy as np

from collections import Counter, defaultdict
from functools import total_ordering
from numbers import Real
from typing import Dict, List, Tuple
from copy import deepcopy

from ..sdp.monomial_utils import (compute_marginal,
                                  name_from_atom_names,
                                  symbol_from_atom_name,
                                  symbol_prod)
from .numbafied import nb_is_do_conditional as is_do_conditional


@total_ordering
class InternalAtomicMonomial(object):
    __slots__ = ["as_lexmon",
                 "as_2d_array",
                 "is_one",
                 "is_zero",
                 "is_do_conditional",
                 "is_knowable",
                 "n_operators",
                 "op_length",
                 "rectified_ndarray",
                 "context",
                 "name",
                 "legacy_name",
                 "is_all_commuting",
                 "signature",
                 "symbol"
                 ]

    def __init__(self, inflation_lp_instance, as_1d_vec: np.ndarray):
        r"""This class models a moment
        :math:`\langle Op_1 Op_2\dots Op_n\rangle` on the inflated problem,
        which cannot be decomposed into products of other moments. It is used
        as a building block for the ``CompoundMonomial`` class. It is
        initialized with a 1D array representing a moment and an instance of
        ``InflationLP``, used for methods that depend on the scenario.


        Parameters
        ----------
        inflation_lp_instance : InflationLP
            An instance of the ``InflationLP`` class. It is used to access
            methods specific to the inflation problem. E.g., when instantiating
            an internal atomic moment, the ``InflationLP`` instance is used to
            check if it already contains such moment.
        array2d : numpy.ndarray
            A moment :math:`\langle Op_1Op_2\dots Op_n\rangle` encoded as a 2D
            array.
        """
        self.context = inflation_lp_instance
        if as_1d_vec.dtype == bool:
            self.as_lexmon = np.flatnonzero(as_1d_vec).astype(np.intc)
        else:
            self.as_lexmon = np.asarray(as_1d_vec, np.intc)

        self.as_2d_array = inflation_lp_instance._lexorder[self.as_lexmon]
        self.n_operators = len(self.as_lexmon)
        self.op_length = inflation_lp_instance._nr_properties
        self.is_one      = (self.n_operators == 0)
        # Hack to account for different meanings of the zero position in the lexorder
        if inflation_lp_instance.problem_type == "lp":
            self.is_zero = False
        elif inflation_lp_instance.problem_type == "sdp":
            self.is_zero = not np.all(self.as_lexmon)
        else:
            raise NotImplementedError("InternalAtomicMonomial requires `lp` or `sdp` parent class.")
        if self.is_one or self.is_zero:
            self.is_all_commuting = True
            self.is_do_conditional = True
            self.is_knowable = True
        else:
            self.is_all_commuting = inflation_lp_instance.all_commuting_q_1d(self.as_lexmon)
            if self.is_all_commuting:
                self.is_do_conditional = is_do_conditional(self.as_2d_array)
            else:
                self.is_do_conditional = False
            if self.is_do_conditional:
                self.is_knowable = self.context._atomic_knowable_q(self.as_2d_array)
            else:
                self.is_knowable = False
        # Save also array with the original setting, not just the effective one
        if self.is_knowable:
            if not (self.is_one or self.is_zero):
                self.rectified_ndarray = np.asarray(
                    inflation_lp_instance.rectify_fake_setting(np.take(self.as_2d_array,
                                                         [0, -2, -1],
                                                         axis=1)),
                    dtype=int)
            else:
                self.rectified_ndarray = np.empty((0,3), dtype=int)

        self.name = self._name
        self.legacy_name = self._raw_name
        self.signature = self._signature
        self.symbol = symbol_from_atom_name(self.name)

    def __copy__(self):
        """Make a copy of the Monomial"""
        cls = self.__class__
        result = cls.__new__(cls)
        for attr in self.__slots__:
            try:
                result.__setattr__(attr, self.__getattribute__(attr))
            except AttributeError:
                pass
        return result

    def __eq__(self, other):
        """Whether the Monomial is equal to the ``other`` Monomial."""
        return self.__hash__() == other.__hash__()

    def __hash__(self):
        """Return the hash of the Monomial."""
        return hash(self.signature)

    def __lt__(self, other):
        """Whether the Monomial is lexicographically smaller than the ``other``
        Monomial.
        """
        return self.signature < other.signature

    def __repr__(self):
        """Return the name of the Monomial."""
        return self.__str__()

    def __str__(self):
        """Return the name of the Monomial."""
        return self.name

    @property
    def as_legacy_lexmon(self):
        return self.as_lexmon

    @property
    def _name(self):
        if self.is_one:
            return "1"
        elif self.is_zero:
            return "0"
        if self.is_do_conditional:
            uncleaned_ops = self.context.InflationProblem._lexrepr_to_dicts[self.as_legacy_lexmon]
            ambigous_outcome_dict = defaultdict(set)
            for op in uncleaned_ops.flat:
                ambigous_outcome_dict[op["Party"]].add(op["Outcome"])
            parties_with_unambigous_outcomes = [(p, s.pop()) for p, s in ambigous_outcome_dict.items() if len(s)==1]
            cleaned_ops = [deepcopy(op) for op in uncleaned_ops.flat]
            for (p, o) in parties_with_unambigous_outcomes:
                for alt_op in cleaned_ops:
                    alt_op["Do Values"] = {k: v for k, v in
                                           alt_op["Do Values"].items()
                                           if not (k==p and o==v)}
            list_of_op_names = [self.context.InflationProblem._interpretation_to_name(op, include_copy_indices=False) for op in cleaned_ops]
        else:
            list_of_op_names = self.context._lexrepr_to_names[self.as_lexmon]
        if self.is_all_commuting:
            return "P[" + " ".join(list_of_op_names) + "]"
        else:
            return "<" + " ".join(list_of_op_names) + ">"

    @property
    def _raw_name(self):
        """A string representing the monomial. In case of knowable monomials,
        it is of the form ``p(outputs|inputs)``. Otherwise it represents the
        expectation value of the monomial with bracket notation.
        """
        if self.is_one:
            return "1"
        elif self.is_zero:
            return "0"
        elif self.is_knowable:
            # Use notation p(outputs|settings)
            # Convention in numpy monomial format is first party = 1
            party_indices = self.rectified_ndarray[:, 0] - 1
            parties       = np.take(self.context.names, party_indices.tolist())
            inputs  = [str(input) for input in self.rectified_ndarray[:, -2]]
            outputs = [str(output) for output in self.rectified_ndarray[:, -1]]
            p_divider = ""
            # We will probably never have more than 1 digit cardinalities,
            # but who knows...
            i_divider = "" if all(len(i) == 1 for i in inputs) else ","
            o_divider = "" if all(len(o) == 1 for o in outputs) else ","
            return ("p" + p_divider.join(parties) +
                    "(" + o_divider.join(outputs) +
                    "|" + i_divider.join(inputs) + ")")
        else:
            return "<" + " ".join(self.context.InflationProblem._lexrepr_to_all_names[
                self.as_legacy_lexmon, 2]) + ">"

    @property
    def _signature(self):
        return (self.n_operators, tuple(self.as_lexmon))

    def compute_marginal(self, prob_array: np.ndarray) -> float:
        """Given a probability distribution, compute the numerical value of
        the Monomial.

        Parameters
        ----------
        prob_array : numpy.ndarray
            The target probability distribution. The dimensions of this array
            are (outcomes_per_party, settings_per_party), where the settings
            are explicitly 1 is there is only one measurement performed.

        Returns
        -------
        float
            The value of the corresponding probability (which can be a marginal
            involving only a few parties)
        """
        if self.is_zero:
            return 0.
        else:
            return compute_marginal(prob_array, self.rectified_ndarray)


class CompoundMoment(object):
    __slots__ = ["as_counter",
                 "factors",
                 "idx",
                 "is_atomic",
                 "is_do_conditional",
                 "is_knowable",
                 "is_one",
                 "is_zero",
                 "knowability_status",
                 "knowable_factors",
                 "n_factors",
                 "n_knowable_factors",
                 "n_operators",
                 "n_unknowable_factors",
                 "name",
                 "legacy_name",
                 "signature",
                 "unknowable_factors",
                 "as_lexmon",
                 "internal_type",
                 "symbol"
                 ]

    def __init__(self, monomials: Tuple[InternalAtomicMonomial]):
        r"""This class models moments :math:`\langle Op_1 Op_2\dots Op_n\rangle
        =\langle Op_i\dots\rangle\langle Op_{i'}\dots\rangle` on the inflated
        problem that are products of other moments. It is built from a tuple of
        instances of the ``InternalAtomicMonomial`` class.

        At initialisation, a moment is classified into knowable, semi-knowable
        or unknowable based on the knowability of each of the atomic moments
        (which in turn is determined through methods of the
        ``InternalAtomicMonomial`` class). This class also computes names for
        the moment, provides the ability to compare (in)equivalence, and to
        assign numerical values to a moment given a probability distribution.

        Parameters
        ----------
        monomials : tuple of InternalAtomicMonomial
            The atomic moments that make up the compound moment.
        """
        self.factors       = tuple(sorted(monomials))
        self.n_factors     = len(self.factors)
        self.n_operators   = sum(factor.n_operators for factor in self.factors)
        self.is_atomic     = (self.n_factors <= 1)
        self.is_do_conditional = all(factor.is_do_conditional for factor in self.factors)
        self.is_knowable   = all(factor.is_knowable for factor in self.factors)
        if self.n_factors == 0:
            self.as_lexmon = np.empty(0, dtype=np.intc)
        else:
            self.as_lexmon = np.hstack([factor.as_lexmon for factor in self.factors])

        knowable_factors   = []
        unknowable_factors = []
        self.as_counter = Counter(self.factors)
        for factor, power in self.as_counter.items():
            if factor.is_knowable:
                knowable_factors.extend([factor] * power)
            else:
                unknowable_factors.extend([factor] * power)
        self.knowable_factors     = tuple(knowable_factors)
        self.unknowable_factors   = tuple(unknowable_factors)
        self.n_knowable_factors   = len(self.knowable_factors)
        self.n_unknowable_factors = len(self.unknowable_factors)
        if self.n_unknowable_factors == 0:
            self.knowability_status = "Knowable"
        elif self.n_unknowable_factors == self.n_factors:
            self.knowability_status = "Unknowable"
        else:
            self.knowability_status = "Semi"
        self.is_one  = (all(factor.is_one for factor in self.factors)
                        or (self.n_factors == 0))
        self.is_zero = any(factor.is_zero for factor in self.factors)

        self.name        = name_from_atom_names([factor.name
                                                 for factor in self.factors])
        self.legacy_name = name_from_atom_names([factor.legacy_name
                                                 for factor in self.factors])
        self.symbol      = symbol_prod([factor.symbol
                                        for factor in self.factors])
        self.signature   = self.factors
        self.internal_type = InternalAtomicMonomial

    def __eq__(self, other):
        """Whether the Monomial is equal to the ``other`` Monomial."""
        if isinstance(other, self.__class__):
            return self.as_counter == other.as_counter
        elif isinstance(other, self.internal_type):
            return (self.n_factors == 1) and other.__eq__(self.factors[0])
        else:
            raise Exception(f"Expected object of class {self.__class__}, received {other}"
                 + f" of class {type(other)}.")
            return False

    def __hash__(self):
        """Return the hash of the Monomial."""
        return hash(self.signature)

    def __len__(self):
        """Return the number of AtomicMonomials in the CompoundMonomial."""
        return self.n_factors

    def __repr__(self):
        """Return the name of the Monomial."""
        return self.__str__()

    def __str__(self):
        """Return the name of the Monomial."""
        return self.name

    def attach_idx(self, idx: int):
        """Assign an index to the Monomial. This is used when generating the
        monomials in a scenario, and identifying them with integers."""
        if idx >= 0:
            self.idx = idx

    def compute_marginal(self, prob_array: np.ndarray) -> float:
        """Given a probability array, evaluate all the knowable atomic moment
        factors in the monomial and return the product of the resulting values.

        The whole moment needs to be knowable (``self.is_knowable = True``),
        else this method will raise an error.

        Parameters
        ----------
        prob_array : np.ndarray
            A conditional probability distribution over the non-inflated
            scenario with dimensions ``(da,db,...,dx,dy,...)`` (i.e., first
            outcomes and then settings for the parties).

        See also
        --------
        CompoundMonomial.evaluate : This function does not rely on the
            knowability of the monomial.
        """
        assert self.is_knowable, ("Only marginals of knowable monomials can " +
                                  f"be computed, and {self} is not knowable.")
        value = 1.
        for factor, power in self.as_counter.items():
            value *= (factor.compute_marginal(prob_array) ** power)
        return value

    def evaluate(self,
                 known_monomials: Dict[InternalAtomicMonomial, float],
                 use_lpi_constraints=True) -> Tuple[float, List, str]:
        """Given a dictionary of values for known atomic monomials,
        substitute all factors of the compound moment that are specified in
        the dictionary with their values and return the product of the values
        and a remainder compound moment made of the remaining factors.

        Parameters
        ----------
        known_monomials : Dict[InternalAtomicMonomial, float]
            A dictionary of known atomic monomials and their values.
        use_lpi_constraints : bool, optional
            Whether compound moments whose factors are partially present in
            ``known_monomials`` are labelled as ``"Semi"`` or ``"Unknowable"``.
            If ``True``, they are labelled as ``"Semi"``. By default ``True``.

        Returns
        -------
        Tuple[float, List, str]
            A tuple where the first element is the value of the knowable parts
            of the compound moment, the second element is a compound monomial
            of all the remaining factors not present in ``known_monomials`` and
            the third element is a string describing whether all factors are
            present in ``known_monomials`` (``"Known"``), some are
            (``"Semi"``), or none are (``"Unknown"``).
        """
        known_value     = 1.
        unknown_counter = Counter()
        for factor, power in self.as_counter.items():
            try:
                known_value *= (known_monomials[factor] ** power)
            except KeyError:
                unknown_counter[factor] = power
        unknown_factors = list(unknown_counter.elements())
        if (len(unknown_factors) == 0):
            known_status = "Known"
        elif ((len(unknown_factors) == self.n_factors)
              or (not use_lpi_constraints)):
            known_status = "Unknown"
        else:
            known_status = "Semi"
            if use_lpi_constraints and isinstance(known_value, Real):
                if np.isclose(known_value, 0):
                    known_status = "Known"
        return known_value, unknown_factors, known_status

    def __copy__(self):
        """Make a copy of the CompoundMonomial"""
        cls = self.__class__
        result = cls.__new__(cls)
        for attr in self.__slots__:
            try:
                result.__setattr__(attr, self.__getattribute__(attr))
            except AttributeError:
                pass
        return result



import platform


from numpy import __version__ as numpy_version
from scipy import __version__ as scipy_version
from sympy import __version__ as sympy_version

import inflation


def about() -> None:
    """Displays information about Inflation, core/optional packages, and
    Python version/platform information.
    """
    try:
        from numba import __version__ as numba_version
    except ImportError:
        numba_version = "Not installed"
    try:
        from mosek import Env
        major, minor, revision = Env.getversion()
        mosek_version = f"{major}.{minor}.{revision}"
    except ImportError:
        mosek_version = "Not installed"

    about_str = f"""
Inflation: Implementations of the Inflation Technique for Causal Inference
==============================================================================
Authored by: Emanuel-Cristian Boghiu, Elie Wolfe and Alejandro Pozas-Kerstjens

Inflation Version:\t{inflation.__version__}

Core Dependencies
-----------------
NumPy Version:\t{numpy_version}
SciPy Version:\t{scipy_version}
SymPy Version:\t{sympy_version}
Numba Version:\t{numba_version}
Mosek Version:\t{mosek_version}

Python Version:\t{platform.python_version()}
Platform Info:\t{platform.system()} ({platform.machine()})
"""
    print(about_str)


if __name__ == "__main__":
    about()



"""
This file contains functions to interact with LP solvers.

@authors: Erica Han, Elie Wolfe
"""

import sys
import mosek
import numpy as np

from typing import List, Dict, Union
from scipy.sparse import coo_array, issparse
from time import perf_counter
from gc import collect
from ..utils import partsextractor, expand_sparse_vec, vstack
from array import array


def drop_zero_rows(coo_mat: coo_array):
    """Drops zero rows from a sparse matrix in place.

    Parameters
    ----------
    coo_mat : coo_array
        Sparse matrix to drop zero rows from.
    """
    if len(coo_mat.shape) == 1:
        coo_mat = coo_mat.reshape((1, coo_mat.shape[0]))
    nz_rows, new_row = np.unique(coo_mat.row, return_inverse=True)
    coo_mat.row = new_row
    coo_mat = coo_mat.reshape((len(nz_rows), coo_mat.shape[1]))
    return coo_mat


def canonical_order(coo_mat: coo_array):
    """Puts a sparse matrix in canonical order in place.

    Parameters
    ----------
    coo_mat : coo_array
        Sparse matrix to put in canonical order.
    """
    if len(coo_mat.shape) == 1:
        coo_mat = coo_mat.reshape((1, coo_mat.shape[0]))
    order = np.lexsort([coo_mat.col, coo_mat.row])
    coo_mat.row = np.asarray(coo_mat.row)[order]
    coo_mat.col = np.asarray(coo_mat.col)[order]
    coo_mat.data = np.asarray(coo_mat.data)[order]
    return coo_mat

def solveLP(objective: Union[coo_array, Dict] = None,
            known_vars: Union[coo_array, Dict] = None,
            semiknown_vars: Dict = None,
            inequalities: Union[coo_array, List[Dict]] = None,
            equalities: Union[coo_array, List[Dict]] = None,
            variables: List = None,
            **kwargs
            ) -> Dict:
    """Wrapper function that converts all dictionaries to sparse matrices to
    pass to the solver.

    Parameters
    ----------
    objective : Union[coo_array, Dict], optional
        Objective function
    known_vars : Union[coo_array, Dict], optional
        Known values of the monomials
    semiknown_vars : Dict, optional
        Semiknown variables
    inequalities : Union[coo_array, List[Dict]], optional
        Inequality constraints
    equalities : Union[coo_array, List[Dict]], optional
        Equality constraints
    variables : List
        Monomials by name in same order as column indices of all other solver
        arguments

    Returns
    -------
    dict
        Primal objective value, dual objective value, problem status, success
        status, dual certificate (as dictionary and sparse matrix), x values,
        and response code.
    """
    # Save solver arguments, unpacking kwargs
    solver_args = locals()
    del solver_args['kwargs']
    solver_args.update(kwargs)

    # Check type for arguments related to the problem
    problem_args = ("objective",
                    "known_vars",
                    "semiknown_vars",
                    "inequalities",
                    "equalities",
                    "lower_bounds",
                    "upper_bounds")
    used_args = {k: v for k, v in solver_args.items()
                 if k in problem_args and v is not None}
    if all(issparse(arg) for arg in used_args.values()):
        assert variables is not None, "Variables must be declared when all " \
                                      "arguments are in sparse matrix form."
    elif all(isinstance(arg, (dict, list)) for arg in used_args.values()):
        if variables is None:
            # Infer variables
            variables = set()
            if objective:
                variables.update(objective)
            if inequalities:
                for ineq in inequalities:
                    variables.update(ineq)
            if equalities:
                for eq in equalities:
                    variables.update(eq)
            if semiknown_vars:
                for x, (c, x2) in semiknown_vars.items():
                    variables.update([x, x2])
            if known_vars:
                variables.update(known_vars)
            variables = sorted(variables)
            solver_args["variables"] = variables
        solver_args.update(convert_dicts(**solver_args))
    else:
        assert variables is not None, "Variables must be declared when " \
                                      "arguments are of mixed form."
        solver_args.update(convert_dicts(**solver_args))
    solver_args.pop("semiknown_vars", None)
    return solveLP_sparse(**solver_args)


blank_coo_array = coo_array((0, 0), dtype=np.int8)
def solveLP_sparse(objective: coo_array = blank_coo_array,
                   known_vars: coo_array = blank_coo_array,
                   inequalities: coo_array = blank_coo_array,
                   equalities: coo_array = blank_coo_array,
                   lower_bounds: coo_array = blank_coo_array,
                   upper_bounds: coo_array = blank_coo_array,
                   solve_dual: bool = False,
                   default_non_negative: bool = True,
                   relax_known_vars: bool = False,
                   relax_inequalities: bool = False,
                   verbose: int = 0,
                   solverparameters: Dict = None,
                   variables: List = None
                   ) -> Dict:
    """Internal function to solve an LP with the Mosek Optimizer API using
    sparse matrices. Columns of each matrix correspond to a fixed order of
    variables in the LP.

    Parameters
    ----------
    objective : coo_array, optional
        Objective function with coefficients as matrix entries.
    known_vars : coo_array, optional
        Known values of the monomials with values as matrix entries.
    inequalities : coo_array, optional
        Inequality constraints in matrix form.
    equalities : coo_array, optional
        Equality constraints in matrix form.
    lower_bounds : coo_array, optional
        Lower bounds of variables with bounds as matrix entries.
    upper_bounds : coo_array, optional
        Upper bounds of variables with bounds as matrix entries.
    solve_dual : bool, optional
        Whether to solve the dual (``True``) or primal (``False``) formulation.
        By default, ``False``.
    default_non_negative : bool, optional
        Whether to set default primal variables as non-negative. By default,
        ``True``.
    relax_known_vars : bool, optional
        Do feasibility as optimization where each known value equality becomes
        two relaxed inequality constraints. E.g., P(A) = 0.7 becomes P(A) +
        lambda >= 0.7 and P(A) - lambda <= 0.7, where lambda is a slack
        variable. By default, ``False``.
    relax_inequalities : bool, optional
        Do feasibility as optimization where each inequality is relaxed by the
        non-negative slack variable lambda. By default, ``False``.
    verbose : int, optional
        Verbosity. Higher means more messages. By default, 0.
    solverparameters : dict, optional
        Parameters to pass to the MOSEK solver. For example, to control whether
        presolve is applied before optimization, set
        ``mosek.iparam.presolve_use`` to ``mosek.presolvemode.on`` or
        ``mosek.presolvemode.off``. Or, control which optimizer is used by
        setting an optimizer type to ``mosek.iparam.optimizer``. See `MOSEK's
        documentation`_ for more details.
    variables : list
        Monomials by name in same order as column indices of all other solver
        arguments

    Returns
    -------
    dict
        Primal objective value, dual objective value, problem status, success
        status, dual certificate (as dictionary and sparse matrix), x values,
        and response code.
    """

    inequalities=drop_zero_rows(inequalities)
    equalities=drop_zero_rows(equalities)
    known_vars=canonical_order(known_vars)
    upper_bounds=canonical_order(upper_bounds)
    lower_bounds=canonical_order(lower_bounds)
    objective=canonical_order(objective)

    if verbose > 1:
        t0 = perf_counter()
        t_total = perf_counter()
        print("Starting pre-processing for the LP solver...")


    if relax_known_vars or relax_inequalities:
        default_non_negative = False

    with mosek.Env() as env:
        with mosek.Task(env) as task:
            # Set parameters for the solver depending on value type
            if solverparameters:
                for param, val in solverparameters.items():
                    if isinstance(val, int):
                        task.putintparam(param, val)
                    elif isinstance(val, float):
                        task.putdouparam(param, val)
                    elif isinstance(val, str):
                        task.putstrparam(param, val)
            task.putintparam(mosek.iparam.sim_reformulation,
                             mosek.simreform.aggressive)
            task.putintparam(mosek.iparam.sim_switch_optimizer, mosek.onoffkey.on)
            task.putintparam(mosek.iparam.optimizer, mosek.optimizertype.new_primal_simplex)
            if solve_dual:
                task.putintparam(mosek.iparam.sim_solve_form,
                                 mosek.solveform.dual)
            else:
                task.putintparam(mosek.iparam.sim_solve_form,
                                 mosek.solveform.primal)
            if verbose > 0:
                # Attach a log stream printer to the task
                task.set_Stream(mosek.streamtype.log, streamprinter)
                task.putintparam(mosek.iparam.log_include_summary,
                                 mosek.onoffkey.on)
                task.putintparam(mosek.iparam.log_storage, 1)
            if verbose < 2:
                task.putintparam(mosek.iparam.log_sim, 0)
                task.putintparam(mosek.iparam.log_intpnt, 0)

            # Initialize constraint matrix
            constraints = vstack((inequalities, equalities))

            nof_primal_inequalities = inequalities.shape[0]
            nof_primal_equalities = equalities.shape[0]

            (nof_primal_constraints, nof_primal_variables) = constraints.shape
            nof_known_vars = known_vars.nnz

            # Initialize b vector (RHS of constraints)
            b = [0] * nof_primal_constraints

            if relax_inequalities:
                # Add slack variable lambda to each inequality
                cons_row = np.hstack(
                    (constraints.row, np.arange(nof_primal_inequalities)))
                cons_col = np.hstack(
                    (constraints.col, np.repeat(nof_primal_variables,
                                                nof_primal_inequalities)))
                cons_data = np.hstack(
                    (constraints.data, np.repeat(1, nof_primal_inequalities)))
                constraints = coo_array((cons_data, (cons_row, cons_col)),
                                         shape=(nof_primal_constraints,
                                                nof_primal_variables + 1))

            if relax_known_vars:
                # Each known value is replaced by two inequalities with slacks
                kv_row = np.tile(
                    np.arange(nof_known_vars * 2),
                    2)
                kv_col = np.hstack((
                    np.tile(known_vars.col, 2),
                    np.broadcast_to(nof_primal_variables, nof_known_vars * 2)
                ))
                kv_data = np.hstack((
                    np.broadcast_to(1, nof_known_vars * 3),
                    np.broadcast_to(-1, nof_known_vars)
                ))

                kv_matrix = coo_array((kv_data, (kv_row, kv_col)),
                                       shape=(nof_known_vars * 2,
                                              nof_primal_variables + 1))
                canonical_order(kv_matrix)
                constraints.resize(*(nof_primal_constraints,
                                     nof_primal_variables + 1))
                b = np.hstack((b, np.tile(known_vars.data, 2)))
            else:
                # Add known values as equalities to the constraint matrix
                kv_matrix = expand_sparse_vec(known_vars)
                b = np.hstack((b, known_vars.data))
                nof_primal_equalities += nof_known_vars

            constraints = vstack((constraints, kv_matrix))
            (nof_primal_constraints, nof_primal_variables) = constraints.shape
            if objective.shape[-1] == 0:
                objective = coo_array((1, nof_primal_variables), dtype=np.int8)

            if verbose > 0:
                print(f"Size of constraint matrix: {constraints.shape}")


            if verbose > 1:
                print("Proceeding with primal initialization...")

            matrix = constraints.tocsc(copy=False)

            if verbose > 1:
                print("Sparse matrix reformat complete...")

            if relax_known_vars or relax_inequalities:
                # Maximize lambda
                # (If maximum slack is still negative, then the unrelaxed
                # LP would be infeasible, whereas positive slack solution
                # implies LP solution strictly interior in the polytope.)
                objective_vector = np.zeros(nof_primal_variables)
                objective_vector[-1] = -1
            else:
                objective_vector = objective.toarray().ravel()

            # Set bound keys and values for constraints
            # Ax >= b where b is 0
            bkc = np.hstack((np.broadcast_to(mosek.boundkey.lo, nof_primal_inequalities),
                             np.broadcast_to(mosek.boundkey.fx,
                                             nof_primal_equalities)
                             ))
            if relax_known_vars:
                bkc = np.hstack((bkc,
                                 np.repeat([mosek.boundkey.lo, mosek.boundkey.up],
                                                 nof_known_vars)
                                 ))
            blc = buc = b

            ub_col = upper_bounds.col
            ub_data = np.zeros(nof_primal_variables)
            ub_data[ub_col] = upper_bounds.data
            lb_col = lower_bounds.col
            lb_data = np.zeros(nof_primal_variables)
            lb_data[lb_col] = lower_bounds.data

            # Set bound keys and bound values for variables
            blx = np.zeros(nof_primal_variables)
            bux = np.zeros(nof_primal_variables)
            ub_col = np.asarray(upper_bounds.col)
            lb_col = np.asarray(lower_bounds.col)
            lb_data = lower_bounds.data
            if default_non_negative:
                bkx = np.repeat(mosek.boundkey.lo, nof_primal_variables)
                bkx[ub_col] = mosek.boundkey.ra
            else:
                bkx = np.repeat(mosek.boundkey.fr, nof_primal_variables)
                bkx[np.setdiff1d(lb_col, ub_col)] = mosek.boundkey.lo
                bkx[np.setdiff1d(ub_col, lb_col)] = mosek.boundkey.up
                bkx[np.intersect1d(ub_col, lb_col)] = mosek.boundkey.ra
            blx[lb_col] = lb_data
            bux[ub_col] = upper_bounds.data


            if relax_known_vars or relax_inequalities:
                bkx[-1] = mosek.boundkey.fr

            # Set the objective sense
            task.putobjsense(mosek.objsense.maximize)

            # Add all the problem data to the task
            numcon = nof_primal_constraints
            numvar = nof_primal_variables
            if verbose > 1:
                print("Starting task.inputdata in Mosek...")
            task.inputdata(# maxnumcon=
                           numcon,
                           # maxnumvar=
                           numvar,
                           # c=
                           array("d", objective_vector),
                           # cfix=
                           0,
                           # aptrb=
                           matrix.indptr[:-1],
                           # aptre=
                           matrix.indptr[1:],
                           # asub=
                           array("i", matrix.indices),
                           # aval=
                           matrix.data,
                           bkc,
                           blc,
                           buc,
                           bkx,
                           blx,
                           bux)
            collect()
            if verbose > 1:
                print("Pre-processing took",
                      format(perf_counter() - t0, ".4f"), "seconds.\n")
                t0 = perf_counter()

            if verbose > 1:
                print("Writing problem to debug_lp.ptf...")
                task.writedata("debug_lp.ptf")

            # Solve the problem
            if verbose > 0:
                print("\nSolving the problem...\n")
            trmcode = task.optimize()
            if verbose > 1:
                print("Solving took", format(perf_counter() - t0, ".4f"),
                      "seconds.")
            basic = mosek.soltype.bas
            (problemsta,
             solutionsta,
             skc,
             skx,
             skn,
             xc,
             xx,
             yy,
             slc,
             suc,
             slx,
             sux,
             snx) = task.getsolution(basic)

            # Get objective values, solutions x, dual values y
            xx = np.asarray(xx, dtype=object)
            yy = np.asarray(yy, dtype=object)

            primal = task.getprimalobj(basic)
            dual = task.getdualobj(basic)
            x_values = dict(zip(variables, xx))
            y_values = yy

            if solutionsta == mosek.solsta.optimal:
                success = True
            else:
                success = False
            status_str = solutionsta.__repr__()
            term_tuple = mosek.Env.getcodedesc(trmcode)
            if solutionsta == mosek.solsta.unknown and verbose > 0:
                print("The solution status is unknown.")
                print(f"   Termination code: {term_tuple}")

            # Extract the certificate as a sparse matrix: y.b - c.x <= 0
            if relax_known_vars:
                y_values = y_values[nof_primal_constraints - nof_known_vars*2:]
            else:
                y_values = y_values[nof_primal_constraints - nof_known_vars:]
            cert_row = [0] * nof_primal_variables
            cert_col = [*range(nof_primal_variables)]
            cert_data = np.zeros((nof_primal_variables,))
            obj_data = objective.toarray().ravel()
            objective_unknown_cols = np.setdiff1d(objective.col, known_vars.col)
            cert_data[objective_unknown_cols] = -obj_data[objective_unknown_cols]
            cert_data[known_vars.col] = y_values[:nof_known_vars] # Assumes known values coded as equalities
            if relax_known_vars:
                cert_data[known_vars.col] += y_values[nof_known_vars:(2*nof_known_vars)]
            sparse_certificate = coo_array((cert_data, (cert_row, cert_col)),
                                            shape=(1, nof_primal_variables))

            # Certificate as a dictionary
            certificate = dict(zip(variables,
                                   sparse_certificate.toarray().ravel().tolist()))

            # Clean entries with coefficient zero
            for x in list(certificate):
                if np.isclose(certificate[x], 0):
                    del certificate[x]

            if verbose > 1:
                print("\nTotal execution time:",
                      format(perf_counter() - t_total, ".4f"), "seconds.")

            return {
                "primal_value": primal,
                "dual_value": dual,
                "status": status_str,
                "success": success,
                "dual_certificate": certificate,
                "sparse_certificate": sparse_certificate,
                "x": x_values,
                "term_code": term_tuple
            }


def streamprinter(text: str) -> None:
    """A stream printer to get output from Mosek.
    
    Parameters
    ----------
    text : str
        Text to print.
    """
    sys.stdout.write(text)
    sys.stdout.flush()


###########################################################################
# ROUTINES RELATED TO SPARSE MATRIX CONVERSION                            #
###########################################################################


def to_sparse(argument: Union[Dict, List[Dict]],
              variables: List) -> coo_array:
    """Convert a solver argument to a sparse matrix to pass to the solver.
    
    Parameters
    ----------
    argument : Union[Dict, List[Dict]]
        Solver argument to convert to sparse matrix.
    variables : List
        Monomials by name in same order as column indices of all other solver
        arguments
    
    Returns
    -------
    coo_array
        Sparse matrix representation of the solver argument.
    """
    if type(argument) == dict:
        var_to_idx = {x: i for i, x in enumerate(variables)}
        data = list(argument.values())
        keys = list(argument.keys())
        col = partsextractor(var_to_idx, keys)
        row = np.zeros(len(col), dtype=int)
        return coo_array((data, (row, col)), shape=(1, len(variables)))
    else:
        # Argument is a list of constraints
        row = []
        for i, cons in enumerate(argument):
            row.extend([i] * len(cons))
        cols = [to_sparse(cons, variables).col for cons in argument]
        col = [c for vec_col in cols for c in vec_col]
        data = [to_sparse(cons, variables).data for cons in argument]
        data = [d for vec_data in data for d in vec_data]
        return coo_array((data, (row, col)),
                          shape=(len(argument), len(variables)))


def convert_dicts(semiknown_vars: Dict = None,
                  variables: List = None,
                  **kwargs) -> Dict:
    """Convert any dictionaries to sparse matrices to send to the solver.
    Semiknown variables are absorbed into the equality constraints.
    
    Parameters
    ----------
    semiknown_vars : Dict, optional
        Semiknown variables
    variables : List
        Monomials by name in same order as column indices of all other solver
        arguments
    
    Returns
    -------
    Dict
        Converted arguments (Unchanged arguments are not returned)
    """

    assert variables is not None, "Variables must be passed."

    args = locals()
    del args['kwargs']
    args.update(kwargs)

    # Arguments converted from dictionaries to sparse matrices
    sparse_args = {k: to_sparse(arg, variables) for k, arg in args.items()
                   if isinstance(arg, (dict, list)) and k != "semiknown_vars"
                   and k != "solverparameters" and k != "variables"}

    # Add semiknown variables to equality constraints
    if semiknown_vars:
        nof_semiknown = len(semiknown_vars)
        nof_variables = len(variables)
        var_to_idx = {x: i for i, x in enumerate(variables)}
        row = np.repeat(np.arange(nof_semiknown), 2)
        col = [(var_to_idx[x], var_to_idx[x2])
               for x, (c, x2) in semiknown_vars.items()]
        col = list(sum(col, ()))
        data = [(1, -c) for x, (c, x2) in semiknown_vars.items()]
        data = list(sum(data, ()))
        semiknown_mat = coo_array((data, (row, col)),
                                   shape=(nof_semiknown, nof_variables))
        if "equalities" in sparse_args:
            sparse_args["equalities"] = vstack(
                (sparse_args["equalities"], semiknown_mat))
        else:
            sparse_args["equalities"] = semiknown_mat
    return sparse_args



"""Inflation
==================
Provides
 1. A tool for describing causal scenarios and reducing them to network form
    (see the definitions of network and non-network causal scenarios in
    arXiv:1707.06476 and arXiv:1909.10519).
 2. A tool for setting up and solving feasibility and optimization problems
    over probability distributions compatible with quantum causal scenarios.
 3. A tool for setting up and solving feasibility and optimization problems
    over probability distributions compatible with theory-independent and
    classical causal scenarios
"""

from .InflationProblem import InflationProblem
from .sdp.InflationSDP import InflationSDP
from .lp.InflationLP import InflationLP
from .optimization_utils import max_within_feasible
from ._about import about
from ._version import __version__

__all__ = ["InflationProblem",
           "InflationSDP",
           "InflationLP",
           "max_within_feasible"]



from setuptools import setup, find_packages

with open('inflation/_version.py') as f:
    exec(f.read())

setup(
    name="inflation",
    version=__version__,
    install_requires=["mosek>=10", "networkx", "numba", "numpy",
                      "scipy>=1.8.0", "sympy", "tqdm"],
    extras_require={
        "docs": ["nbsphinx", "m2r2", "sphinx_rtd_theme", "sphinx_copybutton"]
    },
    author="Emanuel-Cristian Boghiu, Elie Wolfe, Alejandro Pozas-Kerstjens",
    author_email="cristian.boghiu@icfo.eu, ewolfe@pitp.ca, physics@alexpozas.com",
    description="Implementations of the Inflation Technique for Causal Inference",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    python_requires=">=3.9",
    packages=find_packages(exclude=["test", "doc*", "example*"]),
    license="GNU GPL v. 3.0",
    url="https://github.com/ecboghiu/inflation",
    project_urls={"Documentation": "https://ecboghiu.github.io/inflation/_build/html/index.html",
                  "Source": "https://github.com/ecboghiu/inflation",
                  "Issue Tracker": "https://github.com/ecboghiu/inflation/issues"},
    zip_safe=False,  # To avoid problems with Numba, https://github.com/numba/numba/issues/4908
)



"""
This file contains helper functions to manipulate monomials and generate moment
matrices.

@authors: Emanuel-Cristian Boghiu, Elie Wolfe, Alejandro Pozas-Kerstjens
"""
import numpy as np
import sympy

from itertools import permutations, product, combinations_with_replacement
from tqdm import tqdm
from typing import Dict, List, Tuple, Union

from .fast_npa import (apply_source_perm,
                       dot_mon,
                       mon_lexsorted,
                       to_canonical_1d_internal,
                       to_name)
from ..utils import format_permutations


###############################################################################
# FUNCTIONS FOR MONOMIALS                                                     #
###############################################################################

def flatten_symbolic_powers(monomial: sympy.core.symbol.Symbol
                            ) -> List[sympy.core.symbol.Symbol]:
    """If we have powers of a monomial, such as A**3, return a list with
    the factors, [A, A, A].

    Parameters
    ----------
    monomial : sympy.core.symbol.Symbol
        Symbolic monomial, possible with powers.

    Returns
    -------
    List[sympy.core.symbol.Symbol]
        List of all the symbolic factors, with the powers expanded.
    """
    factors          = monomial.as_ordered_factors()
    factors_expanded = []
    for factor in factors:
        base, exp = factor.as_base_exp()
        factors_expanded.extend([base] * exp)
    factors = factors_expanded
    return factors


def reduce_inflation_indices(monomial: np.ndarray) -> np.ndarray:
    """Reduce the inflation indices of a monomial as much as possible. This
    procedure might not give the canonical form directly due to commutations.

    Parameters
    ----------
    monomial : numpy.ndarray
        Input monomial.

    Returns
    -------
    numpy.ndarray
        An equivalent monomial closer to its representative form.
    """
    new_mon    = monomial.copy()
    nr_sources = monomial.shape[1] - 3
    # Pad the monomial with a row a zeros at the front so as to have the
    # relevant inflation indices to begin at 1
    new_mon_padded_transposed = np.concatenate((np.zeros_like(new_mon[:1]),
                                                new_mon)).T
    for source in range(nr_sources):
        copies_used = new_mon_padded_transposed[1 + source]
        _, unique_positions = np.unique(copies_used, return_inverse=True)
        new_mon_padded_transposed[1 + source] = unique_positions
    return new_mon_padded_transposed.T[1:]

def calculate_momentmatrix_1d_internal(cols: List,
                                       notcomm: np.ndarray,
                                       orthomat: np.ndarray,
                                       commuting: bool = False,
                                       verbose: int = 0
                                       ) -> Tuple[np.ndarray, Dict]:
    r"""Calculate the moment matrix. The function takes as input the generating
    set :math:`\{M_i\}_i` encoded as a list of monomials. Each monomial is a
    matrix where each row is an operator and the columns specify the operator
    labels/indices. The moment matrix is the inner product between all possible
    pairs of elements from the generating set. The program outputs the moment
    matrix as a 2d array. Entry :math:`(i,j)` of the moment matrix stores the
    index of the monomial that represents the result of the expectation value
    :math:`\text{Tr}(\rho\cdot M_i^\dagger M_j)` for an unknown quantum state
    :math:`\rho` after applying the substitutions. The program returns the
    moment matrix and the dictionary mapping each monomial in string
    representation to its integer representation.

    Parameters
    ----------
    cols : List
        List of numpy.ndarray representing the generating set in 1d internal format.
    notcomm : numpy.ndarray
        Matrix of commutation relations, given in the format specified by
        `inflation.quantum.fast_npa.commutation_matrix`.
    orthomat : numpy.ndarray
        Matrix of orthogonality relations. Each operator can be identified by an
        integer `i` which also doubles as its lexicographic rank. Given two
        operators with ranks `i`, `j`, ``notcomm[i, j]`` is 1 if the operators
        are orthogonal, and 0 if they do.
    commuting : bool, optional
        Whether the variables in the problem commute or not. By default
        ``False``.
    verbose : int, optional
        How much information to print. By default ``0``.

    Returns
    -------
    Tuple[numpy.ndarray, Dict]
        The moment matrix :math:`\Gamma`, where each entry :math:`(i,j)` stores
        the integer representation of a monomial. The Dict is a mapping from
        string representation to integer representation.
    """
    nrcols = len(cols)
    canonical_mon_to_idx = {}
    momentmatrix = np.zeros((nrcols, nrcols), dtype=np.uint32)
    varidx = 1  # We start from 1 because 0 is reserved for 0
    for (i, mon1), (j, mon2) in tqdm(
            combinations_with_replacement(enumerate(cols), 2),
            disable=not verbose,
            desc="Calculating moment matrix",
            total=int(nrcols * (nrcols + 1) / 2),
    ):
        mon_v1 = to_canonical_1d_internal(dot_mon(mon1, mon2),
                                          notcomm,
                                          orthomat,
                                          commuting=commuting,
                                          apply_only_commutations=False)
        if np.all(mon_v1):  # a zero indicates equal to zero scalar
            if len(mon_v1) > 1 and (not commuting):
                mon_v2 = to_canonical_1d_internal(np.flipud(mon_v1),
                                                  notcomm,
                                                  orthomat,
                                                  commuting=commuting,
                                                  apply_only_commutations=True)
                mon_hash = min(tuple(mon_v1), tuple(mon_v2))
            else:
                mon_hash = tuple(mon_v1)
            try:
                known_varidx = canonical_mon_to_idx[mon_hash]
                momentmatrix[i, j] = known_varidx
                momentmatrix[j, i] = known_varidx
            except KeyError:
                canonical_mon_to_idx[mon_hash] = varidx
                momentmatrix[i, j] = varidx
                momentmatrix[j, i] = varidx
                varidx += 1
    return momentmatrix, canonical_mon_to_idx


###############################################################################
# FUNCTIONS FOR INFLATIONS                                                    #
###############################################################################

def apply_inflation_symmetries(momentmatrix: np.ndarray,
                               inflation_symmetries: np.ndarray,
                               verbose: bool = False
                               ) -> Tuple[np.ndarray,
                                          Dict[int, int],
                                          np.ndarray]:
    """Applies the inflation symmetries, in the form of permutations of the
    rows and colums of a moment matrix, to the moment matrix.

    Parameters
    ----------
    momentmatrix : numpy.ndarray
        The moment matrix.
    inflation_symmetries : numpy.ndarray
        Two-dimensional array where each row represents a permutation of
        the rows and columns of the moment matrix.
    verbose : bool
        Whether information about progress is printed out.

    Returns
    -------
    sym_mm : numpy.ndarray
        The symmetrized version of the moment matrix, where each cell is
        the lowest index of all the Monomials that are equivalent to that
        in the corresponding cell in momentmatrix.
    orbits : Dict[int, int]
        The map from unsymmetrized indices in momentmatrix to their
        symmetrized counterparts in sym_mm.
    repr_values: numpy.ndarray
        An array of unique representative former (unsymmetrized) indices.
        This is later used for hashing indices and making sanitization much
        faster.
    """
    max_value = momentmatrix.max(initial=0)
    if not len(inflation_symmetries):
        repr_values = np.arange(max_value + 1)
        orbits = dict(zip(repr_values, repr_values))
        return momentmatrix, orbits, repr_values
    else:
        old_indices, flat_pos, inverse = np.unique(momentmatrix.ravel(),
                                                   return_index=True,
                                                   return_inverse=True)
        inverse           = inverse.reshape(momentmatrix.shape)
        prev_unique_count = np.inf
        new_unique_count  = old_indices.shape[0]
        new_indices       = np.arange(new_unique_count)
        inversion_tracker = new_indices.copy()
        repr_values = old_indices.copy()
        # We minimize under every element of the inflation symmetry group.
        for permutation in tqdm(inflation_symmetries,
                                disable=not verbose,
                                desc="Applying symmetries      "):
            if prev_unique_count > new_unique_count:
                rows, cols = np.unravel_index(flat_pos, momentmatrix.shape)
            prev_unique_count = new_unique_count
            assert np.array_equal(new_indices, inverse[(rows, cols)]), \
                ("The representatives of the symmetrized indices are " +
                 "not minimal.")
            np.minimum(new_indices,
                       inverse[(permutation[rows], permutation[cols])],
                       out=new_indices)
            unique_values, unique_values_pos, unique_values_inv = \
                np.unique(new_indices,
                          return_index=True,
                          return_inverse=True)
            new_unique_count = unique_values.shape[0]
            if prev_unique_count > new_unique_count:
                inverse           = unique_values_inv[inverse]
                flat_pos          = flat_pos[unique_values_pos]
                repr_values       = repr_values[unique_values_pos]
                inversion_tracker = unique_values_inv[inversion_tracker]
                del unique_values_pos, unique_values_inv
                new_indices = np.arange(new_unique_count)
        prior_min_value = old_indices.min()
        if old_indices.min() != 0:
            new_indices += prior_min_value
        orbits = dict(zip(old_indices, new_indices[inversion_tracker]))
        sym_mm = new_indices[inverse]
        return sym_mm, orbits, repr_values


# TODO @Cristian add this to InflationSDP
def commutation_relations(infSDP):
    """Return a user-friendly representation of the commutation relations.

    Parameters
    ----------
    infSDP : inflation.InflationSDP
        The SDP object for which the commutation relations are to be extracted.

    Returns
    -------
    Tuple[sympy.core.expr.Expr]
        The list of commutators (given as sympy Expressions) that are nonzero.
    """
    from collections import namedtuple
    nonzero = namedtuple("NonZeroExpressions", "exprs")
    data = []
    lexorder = infSDP._lexorder
    lexorder_len = lexorder.shape[0]
    for i in range(lexorder_len):
        for j in range(i, lexorder_len):
            # Most operators commute as they belong to different parties,
            if infSDP._notcomm[i, j] != 0:
                (op1_name, op2_name) = to_name(lexorder[[i, j]],
                                               infSDP.names).split('*')
                op1 = sympy.Symbol(op1_name, commutative=False)
                op2 = sympy.Symbol(op2_name, commutative=False)
                if infSDP.verbose > 0:
                    print(f"{str(op1 * op2 - op2 * op1)} ≠ 0.")
                data.append(op1 * op2 - op2 * op1)
    return nonzero(data)


def construct_normalization_eqs(column_equalities: List[Tuple[int, List[int]]],
                                momentmatrix: np.ndarray,
                                verbose=0,
                                ) -> List[Tuple[int, List[int]]]:
    """Given a list of column level normalization equalities and the moment
    matrix, this function computes the implicit normalization equalities
    between matrix elements. Column-level and monomial-level equalities share
    nearly the same format, they differ merely in whether integers pertain to
    column indices or the indices that represent the unique moment matrix
    elements.

    Parameters
    ----------
    column_equalities : List[Tuple[int, List[int]]]
        The list of equalities between columns in the moment matrix, in the
        form of tuples whose first element is the index of one of the columns,
        and the second element is the list of indices of the columns whose
        corresponding operators sum up to the operator corresponding to the
        first element.
    momentmatrix : numpy.ndarray
        The moment matrix of which the identification between variables shall
        be computed.
    verbose : int, optional
        Verbosity level. By default 0.

    Returns
    -------
    List[Tuple[int, List[int]]]
        The equalities between variables. For each tuple, the first element is
        the index of one of the variables in ``momentmatrix``, and the second
        is the list of variables whose sum corresponds to the first.
    """
    equalities = []
    seen_already = set()
    nof_seen_already = len(seen_already)
    for equality in tqdm(column_equalities,
                         disable=not verbose,
                         desc="Imposing normalization   "):
        for i, row in enumerate(iter(momentmatrix)):
            (normalization_col, summation_cols) = equality
            norm_idx       = row[normalization_col]
            summation_idxs = row[summation_cols]
            if summation_idxs.all():
                summation_idxs.sort()
                seen_already.add(tuple(summation_idxs.flat))
                if len(seen_already) > nof_seen_already:
                    equalities.append((norm_idx, summation_idxs.tolist()))
                    nof_seen_already += 1
    del seen_already
    return equalities


def expand_moment_normalisation(moment: np.ndarray,
                                outcome_cardinalities: List[int],
                                skip_party: List[bool]):
    """Helper function that identifies operators within the monomial that
    correspond to the last outcome, and uses normalisation to produce
    an equality constraint with other monomials. The constraint is expressed as
    `(i, (i1, i2, i3, ...))`, where the moment corresponding to index `i` is
    equal to the sum of moments corresponding to indices
    `(i1, i2, i3, ...)`.

    Parameters
    ----------
    moment : numpy.ndarray
        Moment encoded as a 2D array.
    outcome_cardinalities : List[int]
        List of the cardinalities for the outcomes of the parties.
    skip_party : List[bool]
        Whether each of the parties is considered for normalisation or not.
    """
    eqs = []
    for k, operator in enumerate(moment):
        party = operator[0] - 1
        # Operators that are involved in normalization equalities are
        # those which are unpacked in non-network scenarios
        if (not skip_party[party]
                and operator[-1] == outcome_cardinalities[party] - 2):
            operator_2d = np.expand_dims(operator, axis=0)
            prefix = moment[:k]
            suffix = moment[(k + 1):]
            moments = [moment]
            true_cardinality = outcome_cardinalities[party] - 1
            for outcome in range(true_cardinality - 1):
                variant_operator        = operator_2d.copy()
                variant_operator[0, -1] = outcome
                variant_mon             = np.vstack((prefix,
                                                     variant_operator,
                                                     suffix))
                moments.append(variant_mon)
            if len(moments) == true_cardinality:
                normalization_mon = np.vstack((prefix, suffix))
                eqs.append((normalization_mon, moments))
    return eqs


def lexicographic_order(infSDP) -> Dict[str, int]:
    """Return a user-friendly representation of the lexicographic order.

    Parameters
    ----------
    infSDP : inflation.InflationSDP
        The SDP object for which the commutation relations are to be extracted.

    Returns
    -------
    Dict[str, int]
        The lexicographic order as a dictionary where keys are the monomials in
        the problem and the values are their positions in the lexicographic
        ordering.
    """
    lexorder = {}
    for i, op in enumerate(infSDP._lexorder):
        lexorder[sympy.Symbol(to_name([op], infSDP.names),
                              commutative=False)] = i
    return lexorder



import unittest
import numpy as np

from inflation import InflationProblem, InflationSDP
from inflation.sdp.fast_npa import nb_is_knowable as is_knowable
from inflation.sdp.fast_npa import nb_is_physical as is_physical


scenario = InflationProblem({"lambda": ["A", "B"],
                             "sigma": ["A", "C"],
                             "mu": ["B", "C"]},
                            outcomes_per_party=[2, 2, 2],
                            settings_per_party=[1, 1, 1],
                            inflation_level_per_source=[3, 2, 2],
                            order=('A', 'B', 'C'))

scenario_c = InflationProblem({"lambda": ["A", "B"],
                               "sigma": ["A", "C"],
                               "mu": ["B", "C"]},
                              outcomes_per_party=[2, 2, 2],
                              settings_per_party=[1, 1, 1],
                              inflation_level_per_source=[3, 2, 2],
                              order=('A', 'B', 'C'),
                              classical_sources="all")

class TestKnowable(unittest.TestCase):
    def test_knowable_complete(self):
        A12B23C31 = np.array([[1, 2, 0, 1, 0, 0],
                              [2, 2, 3, 0, 0, 0],
                              [3, 0, 3, 1, 0, 0]])

        self.assertTrue(is_knowable(A12B23C31),
                        "Monomials composed of a complete copy of the scenario"
                        + " are not identified as knowable.")

    def test_knowable_partial(self):
        A11 = np.array([[1, 1, 0, 1, 0, 0]])

        self.assertTrue(is_knowable(A11),
                        "Knowable monomials composed of a subset of parties "
                        + "are not identified as knowable.")

    def test_unknowable_complete(self):
        A11B12C22 = np.array([[1, 1, 0, 1, 0, 0],
                              [2, 1, 2, 0, 0, 0],
                              [3, 0, 2, 2, 0, 0]])

        self.assertFalse(is_knowable(A11B12C22),
                         "Monomials composed of an open copy of the scenario" +
                         " are not identified as unknowable.")

    def test_unknowable_partial(self):
        A11B11C1211 = np.array([[1, 1, 0, 1, 0, 0, 0, 0, 0],
                                [2, 1, 1, 0, 0, 0, 0, 0, 0],
                                [3, 0, 1, 2, 1, 1, 0, 0, 0]])

        self.assertFalse(is_knowable(A11B11C1211),
                         "Unknowable monomials composed of a subset of parties"
                         + " are not identified as unknowable.")


class TestPhysical(unittest.TestCase):
    def test_no_copies(self):
        # Single operator per party and knowable
        monomial = np.array([[1, 0, 1, 1, 0, 0],
                             [2, 1, 0, 1, 0, 0],
                             [3, 1, 1, 0, 0, 0]])
        self.assertEqual(is_physical(monomial), True,
                         "Physical knowable monomials are not identified as " +
                         "physical.")

        # Single operator per party but not knowable
        monomial = np.array([[1, 0, 1, 1, 0, 0],
                             [2, 1, 0, 1, 0, 0],
                             [3, 1, 2, 0, 0, 0]])
        self.assertEqual(is_physical(monomial), True,
                         "Physical unknowable monomials are not identified as "
                         + "physical.")

    def test_two_copies_physical(self):
        monomial = np.array([[1, 0, 1, 1, 0, 0],
                             [1, 0, 2, 2, 0, 0],
                             [2, 1, 0, 1, 0, 0],
                             [3, 2, 1, 0, 0, 0]])
        self.assertEqual(is_physical(monomial), True,
                         "Physical monomials with multiple copies of a party" +
                         " are not identified as such.")

    def test_two_copies_unphysical(self):
        monomial = np.array([[1, 0, 1, 1, 0, 0],
                             [1, 0, 2, 1, 0, 0],
                             [2, 1, 0, 1, 0, 0],
                             [3, 1, 2, 0, 0, 0]])
        self.assertEqual(is_physical(monomial), False,
                         "Unphysical monomials are not identified as " +
                         "unphysical.")


class TestToCanonical(unittest.TestCase):
    sdp = InflationSDP(scenario)

    def test_commutation(self):
        initial = 'A_2_1_0_0_0*A_1_2_0_0_0*B_1_0_1_0_0'
        correct = 'A_1_2_0_0_0*A_2_1_0_0_0*B_1_0_1_0_0'
        self.assertEqual(self.sdp._sanitise_moment(initial),
                         self.sdp._sanitise_moment(correct),
                         "to_canonical has problems with bringing monomials " +
                         "to representative form.")
        
    def test_commutation_with_hybrid_sources(self):
        from inflation.sdp.fast_npa import nb_operators_commute
        
        # Two source scenario. Two operators with overlap on the first source, 
        # with the first source being classical.
        # No intermediate latents.
        pairwise_source_info_both_classical = np.zeros((2, 2, 2), dtype=np.uint8)
        pairwise_source_info_both_quantum = np.zeros((2, 2, 2), dtype=np.uint8)
        pairwise_source_info_first_classical = np.zeros((2, 2, 2), dtype=np.uint8)
        for i in range(2):
            pairwise_source_info_both_classical[i, i] = [1, 1]
            pairwise_source_info_both_quantum[i, i] = [2, 2]
            pairwise_source_info_first_classical[i, i] = [1, 2]
        self.assertFalse(nb_operators_commute(
            np.array([1, 1, 1, 0, 0]),
            np.array([1, 1, 2, 0, 0]),
            sources_to_check_for_pairwise=pairwise_source_info_both_quantum),
            "nb_operators_commute fails to identify " +
            "non-commutativty when overlapping on quantum sources.")
        self.assertTrue(nb_operators_commute(
            np.array([1, 1, 1, 0, 0]),
            np.array([1, 1, 2, 0, 0]),
            sources_to_check_for_pairwise=pairwise_source_info_first_classical),
            "nb_operators_commute fails to identify " +
            "commutativity when overlapping on classical sources.")
        self.assertTrue(nb_operators_commute(
            np.array([1, 1, 1, 0, 0]),
            np.array([1, 1, 2, 1, 0]),
            sources_to_check_for_pairwise=pairwise_source_info_first_classical),
            "nb_operators_commute fails to identify " +
            "commutativity when overlapping on classical sources " +
            "with different settings for the operators.")
        self.assertTrue(nb_operators_commute(
            np.array([1, 1, 1, 0, 0]),
            np.array([1, 1, 1, 1, 0]),
            sources_to_check_for_pairwise=pairwise_source_info_both_classical),
            "nb_operators_commute fails to identify " +
            "commutativity when overlapping on classical sources " +
            "with different settings for the operators.")
        


    def test_ordering_parties(self):
        initial = 'A_1_1_0_0_0*A_1_2_0_0_0*C_0_2_1_0_0*B_1_0_2_0_0'
        correct = 'A_1_1_0_0_0*A_1_2_0_0_0*B_1_0_2_0_0*C_0_2_1_0_0'
        self.assertEqual(self.sdp._sanitise_moment(initial),
                         self.sdp._sanitise_moment(correct),
                         "to_canonical fails to order parties correctly.")


class TestToRepr(unittest.TestCase):
    sdp_commuting    = InflationSDP(scenario_c)
    sdp_noncommuting = InflationSDP(scenario)
    names = sdp_commuting.names

    def test_commuting(self):
        initial = 'A_3_1_0_0_0*A_2_1_0_0_0*A_3_1_0_0_0'
        correct = 'A_1_1_0_0_0*A_2_1_0_0_0'
        self.assertEqual(self.sdp_commuting._sanitise_moment(initial),
                         self.sdp_commuting._sanitise_moment(correct),
                         "Applying commutations to representative form fails.")

    def test_jump_sources(self):
        initial = 'A_3_1_0_0_0*A_2_1_0_0_0*A_3_1_0_0_0'
        correct = 'A_1_1_0_0_0*A_2_1_0_0_0*A_1_1_0_0_0'
        self.assertEqual(self.sdp_noncommuting._sanitise_moment(initial),
                         self.sdp_noncommuting._sanitise_moment(correct),
                         "Skipping inflation indices when computing " +
                         "representatives fails.")

    def test_swap_all_sources(self):
        initial = 'A_1_1_0_0_0*A_2_2_0_0_0*B_2_0_1_0_0'
        correct = 'A_1_1_0_0_0*A_2_2_0_0_0*B_1_0_1_0_0'
        self.assertEqual(self.sdp_noncommuting._sanitise_moment(initial),
                         self.sdp_noncommuting._sanitise_moment(correct),
                         "Swapping all sources and applying factorization " +
                         "commutations fails.")

    def test_swap_single_source(self):
        initial = 'A_1_1_0_0_0*A_2_2_0_0_0*B_1_0_2_0_0*B_2_0_1_0_0'
        correct = 'A_1_1_0_0_0*A_2_2_0_0_0*B_1_0_1_0_0*B_2_0_2_0_0'
        self.assertEqual(self.sdp_noncommuting._sanitise_moment(initial),
                         self.sdp_noncommuting._sanitise_moment(correct),
                         "Swapping a single source fails.")

class TestCommutations(unittest.TestCase):
    scenario = InflationProblem(dag={"s1": ["A", "B", "l1", "l2"],
                                     "s2": ["C", "D"],
                                     "s3": ["l1", "l2"],
                                     "l1": ["C", "D"],
                                     "l2": ["B", "E"]},
                                outcomes_per_party=[2, 2, 2, 2, 2],
                                settings_per_party=[2, 2, 2, 2, 2],
                                inflation_level_per_source=[2, 2, 2],
                                classical_sources=["s2"],
                                nonclassical_intermediate_latents=["l1"],
                                classical_intermediate_latents=["l2"])
    notcomm = scenario._default_notcomm
    meas    = scenario.measurements
    order   = scenario._lexorder
    
    def test_commute_itself(self):
        # All operators should commute with themselves.
        self.assertEqual(self.notcomm.diagonal().sum(),
                         0,
                         "Some operator does not commute with itself.")
    
    def test_same_nonclassical(self):
        # The operators corresponding to different measurements of a same party
        # fed by nonclassical sources do not commute. Take for example A.
        op1_init = np.where(np.all(self.order == self.meas[0][0,0,0], axis=1))
        op1_end  = np.where(np.all(self.order == self.meas[0][0,0,-1], axis=1))
        op2_init = np.where(np.all(self.order == self.meas[0][0,-1,0], axis=1))
        op2_end  = np.where(np.all(self.order == self.meas[0][0,-1,-1], axis=1))
        op1_init, op1_end = op1_init[0].item(), op1_end[0].item()
        op2_init, op2_end = op2_init[0].item(), op2_end[0].item()
        self.assertTrue(
            np.all(self.notcomm[op1_init:op1_end+1, op2_init:op2_end+1]),
            "Operators for different measurements of a same party commute.")
        
    def test_same_classical(self):
        # All operators corresponding to a same party fed by classical sources
        # commute. Take for example E.
        op1_init = np.where(np.all(self.order == self.meas[-1][0,0,0], axis=1))
        op1_end = np.where(np.all(self.order == self.meas[-1][-1,-1,-1], axis=1))
        op1_init, op1_end = op1_init[0].item(), op1_end[0].item()
        self.assertEqual(
            np.sum(self.notcomm[op1_init:op1_end+1, op1_init:op1_end+1]),
            0,
            "Operators of a classical party do not commute.")

    def test_source_parties(self):
        # We restrict to the case of parties not connected by intermediate
        # latents. Regardless of the nature of the sources, they should commute.
        # This applies to all pairs of parties, except BE (connected by a
        # classical latent, tretated later), and CD (connected by a quantum
        # latent and a classical source, so they only commute if the
        # nonclassical copies do not overlap, treated later)
        for pair in [[0, 1], [0, 2], [0, 3], [0, 4],
                     [1, 2], [1, 3], [2, 4], [3, 4]]:
            op1_init = np.where(np.all(self.order == self.meas[pair[0]][0,0,0],
                                       axis=1))
            op1_end = np.where(np.all(self.order == self.meas[pair[0]][-1,-1,-1],
                                      axis=1))
            op2_init = np.where(np.all(self.order == self.meas[pair[1]][0,0,0],
                                       axis=1))
            op2_end = np.where(np.all(self.order == self.meas[pair[1]][-1,-1,-1],
                                      axis=1))
            op1_init, op1_end = op1_init[0].item(), op1_end[0].item()
            op2_init, op2_end = op2_init[0].item(), op2_end[0].item()
            self.assertEqual(self.notcomm[op1_init:op1_end+1,
                                          op2_init:op2_end+1].sum(),
                             0,
                             "Operators for separate parties " + \
                             f"{chr(65+pair[0])} and {chr(65+pair[1])} " + \
                             "do not commute.")
    
    def test_classical_nonclassical_sources(self):
        # The CD case. Here we only have commutation if the nonclassical copies
        # are disjoint
        op1_init = np.where(np.all(self.order == self.meas[2][0,0,0], axis=1))
        op1_end  = np.where(np.all(self.order == self.meas[2][0,-1,-1], axis=1))
        op2_init = np.where(np.all(self.order == self.meas[3][-1,0,0], axis=1))
        op2_end = np.where(np.all(self.order == self.meas[3][-1,-1,-1], axis=1))
        op1_init, op1_end = op1_init[0].item(), op1_end[0].item()
        op2_init, op2_end = op2_init[0].item(), op2_end[0].item()
        self.assertEqual(self.notcomm[op1_init:op1_end+1,
                                      op2_init:op2_end+1].sum(),
                         0,
                         "Operators for separate parties with classical and " \
                         + "nonclassical parents do not commute when " \
                         + "measuring on different nonclassical copies.")

    def test_nonclassical_latent(self):
        # When two parties share a common nonclassical latent ancestor, their
        # operators do not commute only when the quantum copy indices partially
        # overlap. This is the case of CD too.
        op1_init = np.where(np.all(self.order == self.meas[2][0,0,0], axis=1))
        op1_end  = np.where(np.all(self.order == self.meas[2][0,-1,-1], axis=1))
        op2_init = np.where(np.all(self.order == self.meas[3][1,0,0], axis=1))
        op2_end  = np.where(np.all(self.order == self.meas[3][1,-1,-1], axis=1))
        op1_init, op1_end = op1_init[0].item(), op1_end[0].item()
        op2_init, op2_end = op2_init[0].item(), op2_end[0].item()
        self.assertTrue(
            np.all(self.notcomm[op1_init:op1_end+1, op2_init:op2_end+1]),
            "Operators for parties with common nonclassical latent " + \
            "ancestors commute for partially overlapping nonclassical copies.")

    def test_classical_latent(self):
        # When two parties share a common classical latent ancestor, their
        # operators should always commute. This is the case for BE.
        op1_init = np.where(np.all(self.order == self.meas[1][0,0,0], axis=1))
        op1_end = np.where(np.all(self.order == self.meas[1][-1,-1,-1], axis=1))
        op2_init = np.where(np.all(self.order == self.meas[4][0,0,0], axis=1))
        op2_end = np.where(np.all(self.order == self.meas[4][-1,-1,-1], axis=1))
        op1_init, op1_end = op1_init[0].item(), op1_end[0].item()
        op2_init, op2_end = op2_init[0].item(), op2_end[0].item()
        self.assertEqual(
            np.sum(self.notcomm[op1_init:op1_end+1, op2_init:op2_end+1]),
            0,
            "Operators for parties with a common classical latent ancestor " + \
            "do not commute.")